
/*--------------------------------------------------------------------*/
/*--- Simple tool for counting UInstrs, using a C helper.          ---*/
/*---                                                    lk_main.c ---*/
/*--------------------------------------------------------------------*/

/*
   This file is part of Lackey, an example Valgrind tool that does
   some simple program measurement.

   Copyright (C) 2002-2004 Nicholas Nethercote
      njn25@cam.ac.uk

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of the
   License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307, USA.

   The GNU General Public License is contained in the file COPYING.
*/

#include "vg_skin.h"
#include "lk_helper.h"
#include "utility.h"
#include <stdio.h> // for "FILE"
#include "valgrind.h" // for VALGRIND_DISCARD_TRANSLATIONS

extern Int numLoads;
extern Int numStores;

Int main_entry=0;

// DENNIS added these globals
Int valgrindMode = -1;
Bool doOutputTimestamps = False;
Bool doTraceAfterForce = False;
Bool doPdgToTrace = False;

mapping_force * mappingsToForce[MAX_MAPPINGS_TO_FORCE];
Int nextMappingToForceSlot = 0; // this value also equals the number of mappings to force!

force_helper_params *forceHelperParams[MAX_PARAM_STRUCTS_ALLOWED];
Int nextForceHelperParamSlot = 0; // this value also equals the number of force helper param structs!

instr_info *lastInstrExecutedFromMainSourceFile = NULL;
Int lastInstrInstanceExecutedFromMainSourceFile = -1;
Int nextOutputPart = 0;
Addr smallestInstrAddrExecutedFromMainSourceFile = 0xFFFFFFFF;
Addr largestInstrAddrExecutedFromMainSourceFile =  0x00000000;

#define VERIFY_ROOM_FOR_PARAM \
   if(nextForceHelperParamSlot >= MAX_PARAM_STRUCTS_ALLOWED) \
   { \
      VG_(printf)("Error: ran out of room for new force helper parameter structures!\n"); \
      VG_(exit)(0); \
   } \



/* Nb: use ULongs because the numbers can get very big */
#define SHADOW1(tempreg) ((tempreg)+2)
#define VG_IS_FLAG_OVERLAP(set1,set2) \
   (( ((FlagSet)set1) & ((FlagSet)set2) ) != 0 )

typedef 
   enum {
      GETV = DUMMY_FINAL_UOPCODE + 1, 
      PUTV,
      INS,
   }
   MemCheckOpcode;

Addr helper_reg_dep_ports[7]={
    NULL,
    (Addr)&XZ_HELPER_reg_dep_port1,
    (Addr)&XZ_HELPER_reg_dep_port2,
    (Addr)&XZ_HELPER_reg_dep_port3,
    (Addr)&XZ_HELPER_reg_dep_port4,
    (Addr)&XZ_HELPER_reg_dep_port5,
    (Addr)&XZ_HELPER_reg_dep_port6};
#ifdef VALUE_WANTED
Addr helper_trace_values[5]={
     NULL,
    (Addr) & XZ_HELPER_trace_value1,
    (Addr) & XZ_HELPER_trace_value2,
    NULL,
    (Addr) & XZ_HELPER_trace_value4};
#endif
Int open_tid;
Int output_fp, output_fp_pos,output_fp_sim;
extern XZ_shadow_reg sh_regs[8];
extern VgHashTable instr_info_table;
extern VgHashTable fun_entry_table;
extern VgHashTable x86InstrInfoTable;
extern VgHashTable suppressedMem;
extern VgHashTable suppressedArch;
extern VgHashTable suppressedArchS;
extern VgHashTable suppressedTemp;

extern VG_(env_setenv) (Char***, Char*, Char*);

void DJ_processForceFile(FILE *fp, Int numLines)
{
   // read in the input file of mapping(s) to force, and store the information in
   // internal data structures
   Int i, j, k;
   Char c;
   
   Addr addr;
   Int instanceNum;
   UInt definedVal;
   UInt definedFlags;
   Int definedReg;
   UInt usedVal;
   Int usedReg;
   mapping_force *mapping;
   
   char str[128];
   char str2[256];
   
   if(numLines > MAX_MAPPINGS_TO_FORCE)
   {
      VG_(printf)("Error: trying to force %d mappings, but maximum of %d allowed\n", numLines, MAX_MAPPINGS_TO_FORCE);
      VG_(exit(0));
   }
   
Int currLine;
for(currLine = 0; currLine < numLines; currLine++)
{
   if(nextMappingToForceSlot == MAX_MAPPINGS_TO_FORCE)
   {
      VG_(printf)("Error: ran out of room for more mappings to force!");
      VG_(exit)(0);
   }
   
   mappingsToForce[nextMappingToForceSlot] = (mapping_force *)(calloc(1, sizeof(mapping_force)));
   mapping = mappingsToForce[nextMappingToForceSlot];
   nextMappingToForceSlot++;
   
   // initialize currInstance to -1 since the x86 instruction to force hasn't been actually exercised yet
   mapping->currInstance = -1;
   
   // read the address
   memset(&str, '\0', 128);
   str[0] = '0';
   str[1] = 'x';
   for(i = 2; (c = (char)(fgetc(fp))) != ' '; i++)
   {
      str[i] = c;
   }
   sscanf(&str, "%x", &addr);
   mapping->instructionAddr = addr;
   
   // read the instance number
   memset(&str, '\0', 128);
   for(i = 0; (c = (char)(fgetc(fp))) != ' '; i++)
   {
      str[i] = c;
   }
   sscanf(&str, "%i", &instanceNum);
   mapping->instanceNum = instanceNum;
   
   // read the defined value (need to specially handle heap and stack address values here)
   memset(&str2, '\0', 256);
   Bool definesFlag = False;
   Bool isStackAddr = False;
   Bool isHeapAddr = False;
   for(i = 0; (c = (char)(fgetc(fp))) != '!'; i++)
   {
      if(c == FLAG_DELIM)
      {
         definesFlag = True;
      }
      if((c=='_') || (c>='A' && c<='Z') || (c>='a' && c<='z'))
      {
         isStackAddr = True;
      }
      str2[i] = c;
   }
   if(str2[0] == '0' && str2[1] == 'x')
   {
      isStackAddr = False; // may be prematurely set to "True" above when we see the 'x' in the "0x"
      isHeapAddr = True;
   }
   
   if(isStackAddr)
   {
      mapping->definedAddrType = STACK_ADDR;
      
      // read the function name
      memset(&str, '\0', 128);
      j = 0;
      for(i = 0; str2[i] != '#'; i++)
      {
         str[j] = str2[i];
         j++;
      }
      mapping->funcName = (Char *)(calloc( 1, (VG_(strlen)(str)+1)*sizeof(Char) ));
      VG_(strcpy)(mapping->funcName, str);
      
      // read the function instance number
      memset(&str, '\0', 128);
      j = 0;
      i++;
      for(; str2[i] != '/'; i++)
      {
         str[j] = str2[i];
         j++;
      }
      sscanf(&str, "%u", &(mapping->funcInst));
      
      // read the offset
      memset(&str, '\0', 128);
      j = 0;
      i++;
      for(; str2[i] != FLAG_DELIM && str2[i] != '\0'; i++)
      {
         str[j] = str2[i];
         j++;
      }
      sscanf(&str, "%u", &(mapping->definedVal));
   }
   else if(isHeapAddr)
   {
      mapping->definedAddrType = HEAP_ADDR;
      
      // read the heap instruction
      memset(&str, '\0', 128);
      j = 0;
      for(i = 0; str2[i] != '#'; i++)
      {
         str[j] = str2[i];
         j++;
      }
      sscanf(&str, "%x", &(mapping->heapInstr));
      
      // read the chunk number
      memset(&str, '\0', 128);
      j = 0;
      i++;
      for(; str2[i] != '/'; i++)
      {
         str[j] = str2[i];
         j++;
      }
      sscanf(&str, "%u", &(mapping->chunkNum));
      
      // read the offset
      memset(&str, '\0', 128);
      j = 0;
      i++;
      for(; str2[i] != FLAG_DELIM && str2[i] != '\0'; i++)
      {
         str[j] = str2[i];
         j++;
      }
      sscanf(&str, "%u", &(mapping->definedVal));
   }
   else
   {
      mapping->definedAddrType = NOT_ADDR;
      
      // regular number
      memset(&str, '\0', 128);
      j = 0;
      for(i = 0; str2[i] != FLAG_DELIM && str2[i] != '\0'; i++)
      {
         str[j] = str2[i];
         j++;
      }
      sscanf(&str, "%u", &definedVal);
      mapping->definedVal = definedVal;
   }
   
   if(definesFlag == True)
   {  
      mapping->hasDefinedFlags = True;
      
      memset(&str, '\0', 128);
      j = 0;
      i++;
      for(; str2[i] != '\0'; i++)
      {
         str[j] = str2[i];
         j++;
      }
      sscanf(&str, "%u", &definedFlags);
      mapping->definedFlags = definedFlags;
   }
   else
   {
      mapping->hasDefinedFlags = False;
   }
   
   // read the defined register
   c = (char)(fgetc(fp));
   if(c != 'r')
   {
      VG_(printf)("Error reading mappings_to_force file: unexpected character: %c\n", c);
      VG_(exit(0));
   }
   memset(&str, '\0', 128);
   for(i = 0; (c = (char)(fgetc(fp))) != ' '; i++)
   {
      str[i] = c;
   }
   sscanf(&str, "%i", &definedReg);
   mapping->definedReg = definedReg;
   
   // mark the defined value as *not* having been forced yet
   mapping->definedWasForced = False;
   
   // bypass the dollar sign
   c = (char)(fgetc(fp));
   if(c != '$')
   {
      VG_(printf)("Error reading mappings_to_force file: unexpected character #2: %c\n", c);
      VG_(exit(0));
   }

   // from this point on, the number of !'s in the remaining line equals the number of used variables!
   memset(&str2, '\0', 256);
   i = 0;
   c = (char)(fgetc(fp));
   Int numUses = 0;
   while(c != '\n' && c != EOF)
   {
      if(c == '!')
      {
         numUses++;
      }
      str2[i] = c;
      c = (char)(fgetc(fp));
      i++;
   }
   mapping->numUses = numUses;
   str2[i] = ' '; // ensure there's an ending space after the last use
   // at this point, we know the total # of uses, and they've been read into the buffer str2
   if(numUses <= 0)
   {
      // there are no uses here
      mapping->usedVals = NULL;
      mapping->usedRegs = NULL;
      mapping->usedWasForced = NULL;
      mapping->usedLocationWasReturned = NULL;
      mapping->usedAddrType = NULL;
      mapping->heapInstrs = NULL;
      mapping->chunkNums = NULL;
      mapping->funcNames = NULL;
      mapping->funcInsts = NULL;
   }
   else
   {
      // read in the uses (and mark each one as having NOT yet been forced during execution)
      mapping->usedVals = (UInt *)(calloc(1, numUses*sizeof(UInt)));
      mapping->usedRegs = (Int *)(calloc(1, numUses*sizeof(Int)));
      mapping->usedWasForced = (Bool *)(calloc(1, numUses*sizeof(Bool)));
      mapping->usedLocationWasReturned = (Bool *)(calloc(1, numUses*sizeof(Bool)));
      mapping->usedAddrType = (UInt *)(calloc(1, numUses*sizeof(UInt)));
      mapping->heapInstrs = (Addr *)(calloc(1, numUses*sizeof(Addr)));
      mapping->chunkNums = (UInt *)(calloc(1, numUses*sizeof(UInt)));
      mapping->funcNames = (Char **)(calloc(1, numUses*sizeof(Char*)));
      mapping->funcInsts = (UInt *)(calloc(1, numUses*sizeof(UInt)));

      // scan buffer str2 to insert the used values and regsiters into internal data structures
      if(str2[0] != ' ')
      {
         VG_(printf)("Error reading mappings_to_force file: unexpected character #3\n");
         VG_(exit(0));
      }
      Int i = 1; // pointer into str2: skip the first character in str2, which is a space
      Int k = 0; // pointer into str
      for(j = 0; j < numUses; j++)
      {
         // read the next use
         Bool isStackVal = False;
         Bool isHeapVal = False;
         if(str2[i] == '0' && str2[i+1] == 'x')
         {
            isHeapVal = True;
         }
         else if((str2[i]=='_') || (str2[i]>='A' && str2[i]<='Z') || (str2[i]>='a' && str2[i]<='z'))
         {
            isStackVal = True;
         }
         
         if(isStackVal)
         {
            mapping->usedAddrType[j] = STACK_ADDR;
            
            // read func name
            memset(&str, '\0', 128);
            c = str2[i];
            while(c != '#')
            {
               str[k] = c;
               k++;i++;
               c = str2[i];
            }
            mapping->funcNames[j] = (Char *)(calloc( 1, (VG_(strlen)(str)+1)*sizeof(Char) ));
            VG_(strcpy)(mapping->funcNames[j], str);
            
            // read func instance
            i++; k = 0;
            memset(&str, '\0', 128);
            c = str2[i];
            while(c != '/')
            {
               str[k] = c;
               k++;i++;
               c = str2[i];
            }
            sscanf(&str, "%u", &(mapping->funcInsts[j]));
            
            // read offset
            i++; k = 0;
            memset(&str, '\0', 128);
            c = str2[i];
            while(c != '!')
            {
               str[k] = c;
               k++;i++;
               c = str2[i];
            }
            sscanf(&str, "%u", &(mapping->usedVals[j]));
         }
         else if(isHeapVal)
         {
            mapping->usedAddrType[j] = HEAP_ADDR;
            
            // read heap instruction
            memset(&str, '\0', 128);
            c = str2[i];
            while(c != '#')
            {
               str[k] = c;
               k++;i++;
               c = str2[i];
            }
            sscanf(&str, "%x", &(mapping->heapInstrs[j]));
            
            // read chunk number
            i++; k = 0;
            memset(&str, '\0', 128);
            c = str2[i];
            while(c != '/')
            {
               str[k] = c;
               k++;i++;
               c = str2[i];
            }
            
            sscanf(&str, "%u", &(mapping->chunkNums[j]));
            
            // read offset
            i++; k = 0;
            memset(&str, '\0', 128);
            c = str2[i];
            while(c != '!')
            {
               str[k] = c;
               k++;i++;
               c = str2[i];
            }
            sscanf(&str, "%u", &(mapping->usedVals[j]));
         }
         else
         {
            mapping->usedAddrType[j] = NOT_ADDR;
            
            memset(&str, '\0', 128);
            c = str2[i];
            while(c != '!')
            {
               str[k] = c;
               k++;i++;
               c = str2[i];
            }
            sscanf(&str, "%u", &usedVal);
            mapping->usedVals[j] = usedVal;
         }
         
         i++; // next character should be "r"
         if(str2[i] != 'r')
         {
            VG_(printf)("Error reading mappings_to_force file: unexpected character #4\n");
            VG_(exit(0));
         }
         i++;
         memset(&str, '\0', 128);
         k = 0;
         c = str2[i];
         while(c != ' ')
         {
            str[k] = c;
            k++;i++;
            c = str2[i];
         }
         sscanf(&str, "%i", &usedReg);
         
         mapping->usedRegs[j] = usedReg;
         mapping->usedWasForced[j] = False;
         mapping->usedLocationWasReturned[j] = False;
         
         i++; // ready to read the next character (if there's another use to process)
         k = 0;
      }
   }
   
   /* For debugging purposes */
   /*
   printf("LINE #%d of %d\n", currLine+1, numLines);
   printf("================\n");
   printf("addr: 0x%x\n", mapping->instructionAddr);
   printf("instance: %i\n", mapping->instanceNum);
   
   if(mapping->definedAddrType == NOT_ADDR)
   {
       if(mapping->hasDefinedFlags == True)
       {
          printf("defines: %u_%u(r%d)\n", mapping->definedVal, mapping->definedFlags, mapping->definedReg);
       }
       else
       {
          printf("defines: %u(r%d)\n", mapping->definedVal, mapping->definedReg);
       }
   }
   else if(mapping->definedAddrType == STACK_ADDR)
   {
       if(mapping->hasDefinedFlags == True)
       {
          printf("defines: %s#%u/%u_%u(r%d)\n", mapping->funcName, mapping->funcInst, mapping->definedVal, mapping->definedFlags, mapping->definedReg);
       }
       else
       {
          printf("defines: %s#%u/%u(r%d)\n", mapping->funcName, mapping->funcInst, mapping->definedVal, mapping->definedReg);
       }
   }
   else if(mapping->definedAddrType == HEAP_ADDR)
   {
       if(mapping->hasDefinedFlags == True)
       {
          printf("defines: %x#%u/%u_%u(r%d)\n", mapping->heapInstr, mapping->chunkNum, mapping->definedVal, mapping->definedFlags, mapping->definedReg);
       }
       else
       {
          printf("defines: %x#%u/%u(r%d)\n", mapping->heapInstr, mapping->chunkNum, mapping->definedVal, mapping->definedReg);
       }
   }
   
   printf("# uses: %d\n", mapping->numUses);
   for(i = 0; i < mapping->numUses; i++)
   {
      if(mapping->usedAddrType[i] == NOT_ADDR)
      {
         printf("   use #%d: %u(r%d)\n", (i+1), mapping->usedVals[i], mapping->usedRegs[i]);
      }
      else if(mapping->usedAddrType[i] == STACK_ADDR)
      {
         printf("   use #%d: %s#%u/%u(r%d)\n", (i+1), mapping->funcNames[i], mapping->funcInsts[i], mapping->usedVals[i], mapping->usedRegs[i]);
      }
      else if(mapping->usedAddrType[i] == HEAP_ADDR)
      {
         printf("   use #%d: %x#%u/%u(r%d)\n", (i+1), mapping->heapInstrs[i], mapping->chunkNums[i], mapping->usedVals[i], mapping->usedRegs[i]);
      }
   }
   
   exit(0);
   */
   
} // end loop to read in each mapping line

} // end function DJ_processForceFile

void SK_(pre_clo_init)(void)
{
   // Delete any "last pid" file that might be in existence.  It will be re-created after
   // execution is complete (in case any errors occur before that, then the file will NOT
   // be created, and that's how one can tell whether execution terminated successfully or not
   unlink("#vg_last_pid.txt");
   
   unlink("#vg_stmt_trace.txt");
   
   unlink("#vg_start_record_stmts.txt");

   // DENNIS added this to set the global flag indicating which mode valgrind is in
   valgrindMode = VG_(getValgrindMode)();
   
   
   // DENNIS added this to set the global flag indicating whether valgrind should
   // output timestamp values or not in the PDG file (only valid in PDG-creation mode)
   if(VG_(doTimestampPdg)() == False)
   {
      // *not* outputting timestamps
      doOutputTimestamps = False;
   }
   else
   {
      // Valgrind should output timestamps in the PDG file.
      doOutputTimestamps = True;
      
      // if we're not running in pdg-creation mode, then the timestamp-pdg flag doesn't
      // apply here, so output an informative error message
      if(valgrindMode != MODE_PDG_CREATION)
      {
         VG_(printf)("Error: timestamp-pdg flag specified,\n");
         VG_(printf)("   but timestamp-pdg flag is only valid in PDG-CREATION mode.\n");
         VG_(exit)(0);
      }
   }
   
   // DENNIS added this to set the global flag indicating whether valgrind should
   // trace used/defined value mappings in instructions, even in force-mapping mode
   // (only valid in force-mapping mode)
   if(VG_(doTraceAfterForce)() == False)
   {
      // *not* tracing value mapping when forcing
      doTraceAfterForce = False;
   }
   else
   {
      // Valgrind should trace value mappings in addition to doing the force
      doTraceAfterForce = True;
      
      // in this mode, timestamps *need* to be recorded so that a trace can be created later
      doOutputTimestamps = True;
      
      // if we're not running in force-mapping mode, then the trace-after-force flag doesn't
      // apply here, so output an informative error message
      if(valgrindMode != MODE_FORCE_MAPPING)
      {
         VG_(printf)("Error: trace-after-force flag specified,\n");
         VG_(printf)("   but trace-after-force flag is only valid in FORCE-MAPPING mode.\n");
         VG_(exit)(0);
      }
      
      #ifdef OPTIMIZE_TRACE_AFTER_FORCE
      // read in the text file of instrs/mappings to keep in the computed trace
      readMappingsToKeep();
      #endif
   }
   
   // DENNIS added this to set the global flag indicating whether valgrind should
   // internally convert the PDG info to TRACE info prior to outputting it
   // (only valid in pdg-creation mode and force-mapping mode)
   if(VG_(doPdgToTrace)() == False)
   {
      // *not* converting PDG to TRACE before outputting
      doPdgToTrace = False;
   }
   else
   {
      // Valgrind should convert PDG to TRACE before outputting it
      doPdgToTrace = True;
      
      // in this mode, timestamps *need* to be recorded so that a trace can be created
      doOutputTimestamps = True;
      
      // if we're not running in one of the appropriate modes, then the pdg-to-trace flag doesn't
      // apply here, so output an informative error message
      if(valgrindMode != MODE_PDG_CREATION && valgrindMode != MODE_FORCE_MAPPING)
      {
         VG_(printf)("Error: pdg-to-trace flag specified,\n");
         VG_(printf)("   but pdg-to-trace flag is only valid in PDG-CREATION and FORCE-MAPPING modes.\n");
         VG_(exit)(0);
      }
   }
   
   
   if(valgrindMode == MODE_FORCE_MAPPING)
   {
      // assume there exists an input file called "#vg_to_force.txt" that contains a list
      // of mappings to force during this program execution.  Read in that file here.  If
      // the file doesn't exist, emit an error message and terminate valgrind
      FILE *fp = fopen("#vg_to_force.txt", "r");
      if(fp == NULL)
      {
         VG_(printf)("Error: valgrind was specified to run in FORCE-MAPPING mode,\n");
         VG_(printf)("   but the input file \"#vg_to_force.txt\" doesn't exist.\n");
         VG_(exit)(0);
      }
      
      // first count the number of lines in the input file
      Char c;
      Int numLines = 0;
      for( ; (c = (char)(fgetc(fp))) != EOF; )
      {
         if(c == '\n')
         {
            numLines++;
         }
      }
      fclose(fp);
      fp = fopen("#vg_to_force.txt", "r");
      
      DJ_processForceFile(fp, numLines);
      
      fclose(fp);
   }
   
   if(valgrindMode == MODE_SUPPRESS)
   {
      // read in the file containing the initial suppression points (if specified)
      readSuppressionFile();
   }

   fprintf(stderr, "***********************************\n");
   switch(valgrindMode)
   {
      case MODE_PDG_CREATION:
      {
         fprintf(stderr, "VALGRIND PDG-CREATION MODE");
         if(doOutputTimestamps == True)
         {
            fprintf(stderr, " (with timestamps)");
         }
         fprintf(stderr, "\n");
         break;
      }
      case MODE_FORCE_MAPPING:
      {
         fprintf(stderr, "VALGRIND FORCE-MAPPING MODE");
         if(doTraceAfterForce == True)
         {
            fprintf(stderr, " (with trace-after-force)");
         }
         fprintf(stderr, "\n");
         break;
      }
      case MODE_STMT_COVERAGE:
      {
         fprintf(stderr, "VALGRIND STMT-COVERAGE MODE\n");
         break;
      }
      case MODE_SUPPRESS:
      {
         fprintf(stderr, "VALGRIND SUPPRESSION MODE!\n");
         break;
      }
      default:
      {
         fprintf(stderr, "VALGRIND UNKNOWN MODE!! HELP! HELP!\n");
         break;
      }
   }
   fprintf(stderr, "***********************************\n\n");

   VG_(details_name)            ("Dynamic Dependence Graph");
   VG_(details_version)         (NULL);
   VG_(details_description)     ("a Valgrind tool");
   VG_(details_copyright_author)(
      "Copyright (C) 2005-2007, and GNU GPL'd, by Xiangyu Zhang.  Modified by Dennis Jeffrey.");
   VG_(details_bug_reports_to)  (VG_BUGS_TO);
   VG_(details_avg_translation_sizeB) ( 228 );

   VG_(needs_shadow_regs)         ();
   VG_(needs_extended_UCode)      ();
   VG_(needs_syscall_wrapper)     ();
   VG_(needs_shadow_memory)       ();

   open_tid=VG_(gettid)();
   
   // DENNIS changed this so that these files are not created (I'm not using them right now)
   /*
   output_fp_sim= VG_(open)("SLICING.sim_output", VKI_O_CREAT|VKI_O_TRUNC|VKI_O_WRONLY,
                                       VKI_S_IRUSR|VKI_S_IWUSR);
   output_fp= VG_(open)("SLICING.output", VKI_O_CREAT|VKI_O_TRUNC|VKI_O_WRONLY,
                                       VKI_S_IRUSR|VKI_S_IWUSR);
   output_fp_pos= VG_(open)("SLICING.output_pos", VKI_O_CREAT|VKI_O_TRUNC|VKI_O_WRONLY,
                                       VKI_S_IRUSR|VKI_S_IWUSR);
   */
   output_fp_sim= VG_(open)("/dev/null", VKI_O_CREAT|VKI_O_TRUNC|VKI_O_WRONLY,
                                       VKI_S_IRUSR|VKI_S_IWUSR);
   output_fp= VG_(open)("/dev/null", VKI_O_CREAT|VKI_O_TRUNC|VKI_O_WRONLY,
                                       VKI_S_IRUSR|VKI_S_IWUSR);
   output_fp_pos= VG_(open)("/dev/null", VKI_O_CREAT|VKI_O_TRUNC|VKI_O_WRONLY,
                                       VKI_S_IRUSR|VKI_S_IWUSR);
   // END DENNIS change
   
   
   if (output_fp < 0 || output_fp_pos < 0 || output_fp_sim<0) {
      VG_(message)(Vg_UserMsg, "FATAL: can't open output file");
      VG_(exit)(1);
   }

   VG_(XZ_shadow_num)=3;
   init_shadow_memory();
   instr_info_table=VG_(HT_construct)();
   fun_entry_table=VG_(HT_construct)();
   x86InstrInfoTable = VG_(HT_construct)();
   suppressedMem = VG_(HT_construct)();
   suppressedArch = VG_(HT_construct)();
   suppressedArchS = VG_(HT_construct)();
   suppressedTemp = VG_(HT_construct)();   

   VG_(register_compact_helper) ((Addr) & XZ_HELPER_get_inst);
   VG_(register_compact_helper) ((Addr) & XZ_HELPER_increment_inst);
   VG_(register_compact_helper) ((Addr) & XZ_HELPER_setup_CD);
   VG_(register_compact_helper) ((Addr) & XZ_HELPER_reg_dep_port1);
   VG_(register_compact_helper) ((Addr) & XZ_HELPER_reg_dep_port2);
   VG_(register_compact_helper) ((Addr) & XZ_HELPER_load_mem_dep);
   VG_(register_compact_helper) ((Addr) & XZ_HELPER_store_mem_dep);
   VG_(register_compact_helper) ((Addr) & XZ_HELPER_get_timestamp);
   
   VG_(register_noncompact_helper) ((Addr) & DJ_HELPER_increment_force_inst);
   VG_(register_noncompact_helper) ((Addr) & DJ_HELPER_set_flag_at_forced_inst);
   VG_(register_noncompact_helper) ((Addr) & DJ_HELPER_get_original_or_forced_value);
   VG_(register_noncompact_helper) ((Addr) & DJ_HELPER_get_original_or_forced_flags);
   VG_(register_noncompact_helper) ((Addr) & DJ_HELPER_get_jump_addr);
   VG_(register_noncompact_helper) ((Addr) & DJ_set_flags_defined);
   VG_(register_noncompact_helper) ((Addr) & DJ_junk);
   VG_(register_noncompact_helper) ((Addr) & DJ_update_last_instr_main_file);
   VG_(register_noncompact_helper) ((Addr) & DJ_got_func_ret);
   VG_(register_noncompact_helper) ((Addr) & DJ_got_func_entry);
   VG_(register_noncompact_helper) ((Addr) & DJ_HELPER_record_stmt);
   VG_(register_noncompact_helper) ((Addr) & DJ_processTraceEntry);
   
   VG_(register_noncompact_helper) ((Addr) & DJ_debug_trace);
   VG_(register_noncompact_helper) ((Addr) & DJ_print_value);
   VG_(register_noncompact_helper) ((Addr) & DJ_handle_initial_suppression_point);
   VG_(register_noncompact_helper) ((Addr) & DJ_handle_load_to_possibly_suppress);
   VG_(register_noncompact_helper) ((Addr) & DJ_set_flag_for_store_if_suppressed);
   VG_(register_noncompact_helper) ((Addr) & DJ_set_dest_reg_for_load);   
   VG_(register_noncompact_helper) ((Addr) & DJ_mark_not_suppressed_temp);
   VG_(register_noncompact_helper) ((Addr) & DJ_transfer_mark_temp_to_archS);
   VG_(register_noncompact_helper) ((Addr) & DJ_transfer_mark_archS_to_temp);
   VG_(register_noncompact_helper) ((Addr) & DJ_transfer_mark_arch_to_temp);
   VG_(register_noncompact_helper) ((Addr) & DJ_transfer_mark_temp_to_arch);
   VG_(register_noncompact_helper) ((Addr) & DJ_transfer_mark_temp_to_temp);
   VG_(register_noncompact_helper) ((Addr) & DJ_transfer_mark_temp_temp_to_temp);
   VG_(register_noncompact_helper) ((Addr) & DJ_transfer_mark_arch_temp_to_temp);
   
   VG_(register_noncompact_helper) ((Addr) & DJ_record_stmt_info);
   

#ifdef TIMESTAMPED 
   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_trace_timestamp);
   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_trace_control_flow);
#endif

#ifdef VALUE_WANTED
   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_trace_value1);
   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_trace_value2);
   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_trace_value4);
#endif
#ifdef TRACK_ARGV
   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_track_argv);
#endif

#ifdef INTER_CD
   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_create_dummy_CD);
#endif


   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_debug);
   //VG_(register_noncompact_helper)((Addr) & XZ_HELPER_debug_get);
   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_debug_put);
   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_debug_print_reg);
   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_debug_terminate);
   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_debug_BB);
   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_debug_EBP_set_value);
   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_debug_EBP_get_value);
   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_debug_load_value);
   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_debug_store_value);


   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_setup_CD_new);
   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_init_lvs_new);
   //VG_(register_noncompact_helper)((Addr) & XZ_HELPER_init_lvs);

   VG_(register_noncompact_helper) ((Addr) & XZ_HELPER_pred_dep);
   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_reg_dep_port3);
   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_reg_dep_port4);
   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_reg_dep_port5);
   VG_(register_noncompact_helper)((Addr) & XZ_HELPER_reg_dep_port6);
/*
#define MUTT
#ifdef MUTT
   Char ** envp= VG_(client_envp);
   VG_(env_setenv) (&envp, "LD_LIBRARY_PATH", "/scratch/xyzhang/toolchain-i386-glibc/lib");
#endif
*/   
//#ifdef CONTROL_DEP
   // DENNIS commented out
   // read_cd_fun_entries("./cd/CONTROL_DEP");
//#endif
}

void SK_(post_clo_init)(void)
{
}

//Search for the predicate in a basic block(for control dependence)
static Int search_for_predicate(UCodeBlock * cb,int start_point)
{
           Int i;
           UInstr * instr;
           UInstr * cjmp=NULL;
           //for (i=VG_(get_num_instrs) (cb)-1;i>=0; i--) {
           for (i=start_point; i<VG_(get_num_instrs)(cb);i++) {
                  instr=VG_(get_instr) (cb,i);
                  //uFlagsRWU(cb, FlagsOSZACP, FlagsEmpty, FlagsEmpty)
                  //conditional branch
                  if (instr->opcode==JMP && instr->cond!=CondAlways&&instr->jmpkind!=JmpRet ) {
                          cjmp=instr;
                          break;
                  }
           }
           if (!cjmp) return -1;
           for (;i>=0;i--) {
                  instr=VG_(get_instr)(cb,i);
                  //if (instr->flags_w==FlagsOSZACP) {
                  if (VG_IS_FLAG_OVERLAP(instr->flags_w,cjmp->flags_r)) {
                          return i;
                  }
           }
           //print_cb(cb);
           //VG_(skin_panic)("Conditional branch found, however, predicate could not be found");
           return -1;
}

//set the ii and inst of the current instruction to the shadow register of any definition port
static __inline__ void XZ_set_instr_inst(UCodeBlock * cb, instr_info* ii, Int s1, Int s2)
{
   
     //uInstr0(cb,INS,0);
     uInstr2(cb, MOV,   4, Literal, 0, TempReg, s1);
     uLiteral(cb, (UInt) ii);

     
     //uInstr0(cb,INS,0);
     VG_(ccall_R_R) (cb, (Addr) & XZ_HELPER_get_inst, s1, s2, 1);
}


static void synth_XZ_getv (Int arch, Int reg_ii, Int reg_inst)
{
  Int REG;
  for (REG=R_EAX;REG<=R_EDI;REG++) {
      if (REG!=reg_ii && REG!=reg_inst) break;
  }
  VG_(emit_pushv_reg) (4,REG);
  VG_(emit_movv_lit_reg) (4, (UInt)sh_regs, REG);

  VG_(emit_movv_offregmem_reg)(4,8*arch,REG,reg_ii);
  VG_(emit_movv_offregmem_reg)(4,8*arch+4,REG,reg_inst);

  VG_(emit_popv_reg)(4,REG);

}

static void synth_XZ_putv (Int arch, Int reg_ii, Int reg_inst)
{
  Int REG;
  for (REG=R_EAX;REG<=R_EDI;REG++) {
      if (REG!=reg_ii && REG!=reg_inst) break;
  }

  VG_(emit_pushv_reg) (4,REG);
  VG_(emit_movv_lit_reg) (4, (UInt)sh_regs, REG);

  VG_(emit_movv_reg_offregmem)(4,reg_ii,8*arch,REG);
  VG_(emit_movv_reg_offregmem)(4,reg_inst,8*arch+4,REG);

  VG_(emit_popv_reg)(4,REG);
}

void SK_(emit_XUInstr) ( UInstr* u, RRegSet regs_live_before )
{
     switch(u->opcode) {
        case GETV:
            sk_assert(u->tag1==ArchReg&&u->tag2==RealReg && u->tag3==RealReg);
	    synth_XZ_getv(u->val1, u->val2, u->val3);
	    break;
        case PUTV:
            sk_assert(u->tag3==ArchReg&&u->tag1==RealReg && u->tag2==RealReg);
	    synth_XZ_putv(u->val3, u->val1, u->val2);
	    break;
    }
}

Bool SK_(sane_XUInstr)(Bool beforeRA, Bool beforeLiveness, UInstr* u)
{
#  define LIT0 (u->lit32 == 0)
#  define LITm (u->tag1 == Literal ? True : LIT0 )
#  define CC0 (u->flags_r == FlagsEmpty && u->flags_w == FlagsEmpty)
#  define SZi (u->size == 4 || u->size == 2 || u->size == 1)
#  define A1  (u->tag1 == ArchReg)
#  define A2  (u->tag2 == ArchReg)
#  define A3  (u->tag3 == ArchReg)
#  define TR1 (beforeRA ? (u->tag1 == TempReg) : (u->tag1 == RealReg))
#  define TR2 (beforeRA ? (u->tag2 == TempReg) : (u->tag2 == RealReg))
#  define TR3 (beforeRA ? (u->tag3 == TempReg) : (u->tag3 == RealReg))

#  define COND0    (u->cond         == 0)
#  define EXTRA4b0 (u->extra4b      == 0)
#  define SG_WD0   (u->signed_widen == 0)
#  define JMPKIND0 (u->jmpkind      == 0)
#  define CCALL0   (u->argc==0 && u->regparms_n==0 && u->has_ret_val==0 && \
                    ( beforeLiveness                                       \
                    ? u->regs_live_after == ALL_RREGS_LIVE                 \
                    : True ))
#  define XOTHER   (COND0 && EXTRA4b0 && SG_WD0 && JMPKIND0 && CCALL0)

    switch(u->opcode)
    {
	   case GETV: return LIT0 && SZi && CC0 && A1 && TR2 && TR3 && XOTHER;
	   case PUTV: return LIT0 && SZi && CC0 && TR1 && TR2 && A3 && XOTHER;
       case INS: return True;
       case DJ_LEAVE: return True;
       case DJ_RET: return True;
       default:
          VG_(printf)("XZ unhandled opcode: %u\n", u->opcode);
          VG_(skin_panic)("SK_(sane_XUInstr): XZ unhandled opcode");
   }
}

Char* SK_(name_XUOpcode)(Opcode opc)
{
   switch (opc) {
      case GETV:    return "GETV";
      case PUTV:    return "PUTV";
      case INS: return "INS";
      case DJ_LEAVE: return "DJ_LEAVE";
      case DJ_RET: return "DJ_RET";
      default:
         VG_(printf)("XZ unhandled opcode: %u\n", opc);
         VG_(skin_panic)("SK_(name_XUOpcode): unhandled case");
   }
}

void SK_(pp_XUInstr)(UInstr* u)
{
   switch (u->opcode) {
      case GETV: case PUTV:
         VG_(pp_UOperand)(u, 1, u->opcode==PUTV ? 4 : u->size, False);
         VG_(printf)(", ");
         VG_(pp_UOperand)(u, 2, u->opcode==GETV ? 4 : u->size, False);
         VG_(printf)(", ");
         VG_(pp_UOperand)(u, 3, u->opcode==GETV ? 4 : u->size, False);
         break;

      case INS:
      case DJ_LEAVE: 
      case DJ_RET: 
         break;
      default:
         VG_(printf)("XZ unhandled opcode: %u\n", u->opcode);
         VG_(skin_panic)("SK_(pp_XUInstr): XZ unhandled opcode");
   }

}


Int SK_(get_Xreg_usage)(UInstr* u, Tag tag, Int* regs, Bool* isWrites)
{
#  define RD(ono)    VG_UINSTR_READS_REG(ono, regs, isWrites)
#  define WR(ono)    VG_UINSTR_WRITES_REG(ono, regs, isWrites)

   Int n = 0;
   switch (u->opcode) {
      case GETV:    WR(2); WR(3);               break;
      case PUTV:    RD(1); RD(2);               break;
      case INS:                                 break;
      case DJ_LEAVE:                            break;
      case DJ_RET:                              break;

      default:
         VG_(printf)("XZ unhandled opcode: %u\n", u->opcode);
         VG_(skin_panic)("SK_(get_Xreg_usage): XZ unhandled opcode");
   }
   return n;

#  undef RD
#  undef WR
}


//To allocate space for CD, call when the function entry BB is instrumented.
static Int stack_alloc_space_find_index_new(UCodeBlock * cb_in, fun_entry * fe)
{
    Int i,j;
    UInstr * u_in, *tmp_in;
  
    for (i=0;i<VG_(get_num_instrs)(cb_in)-1;i++) {
       u_in = VG_(get_instr)(cb_in, i);
       if (u_in->opcode!=PUT) continue;
       if (u_in->val2==R_EBP) return i;// move %esp, %ebp 
    }//for
    return -1;
}

//mov %esp, %ebp
static void stack_alloc_space_instrument_new(UCodeBlock * cb, UInstr * u_in, fun_entry * fe)
{
                            //uInstr0(cb,INS,0);
                            uInstr2(cb, SUB, 4, Literal, 0, TempReg, u_in->val1);
                            //uLiteral(cb, 32); 
                            uLiteral(cb,fe->alloc_lv * PER_CD_SPACE+LAST_PREDICATE);
                            uInstr2(cb,PUT,4, TempReg, u_in->val1, ArchReg, R_ESP);
                            //uLiteral(cb,fe->alloc_lv * PER_CD_SPACE);

                            //VG_(copy_UInstr)(cb, u_in);

                            //ESP should be changed before initilization becuase ESP change will allocate the page, or else you could get segment fault.        
                            //init the allocated lvs to 0
                            //uInstr0(cb,INS,0);

                            //uInstr2(cb,GET, 4, ArchReg, R_EBP, TempReg, t_reg);

			    VG_(ccall_RL_0) (cb, (Addr) & XZ_HELPER_init_lvs_new, u_in->val1, (UInt)fe, 2);

}

static void stack_alloc_space_instrument_after_new (UCodeBlock * cb, UInstr * u_in, fun_entry * fe)
{
     uInstr2(cb, ADD, 4, Literal, 0, TempReg, u_in->val1);
     uLiteral(cb,fe->alloc_lv * PER_CD_SPACE+LAST_PREDICATE);
}

//find pop %ebp
static Int stack_free_space_find_index(UCodeBlock * cb_in)
{
     Int i,t;
     UInstr * u_in;
     t=-1;
   
     //find there is a return
     for (i=VG_(get_num_instrs) (cb_in)-1;i>=0;i--) {
        u_in=VG_(get_instr) (cb_in,i);
        if (u_in->opcode==JMP ) {
	/*get wrong for __int_malloc because the JMP is not necessary ret*/
        //if (u_in->opcode==JMP && u_in->jmpkind==JmpRet) {
	   t=i;
	   break;
	}
     }
     //VG_(printf) ("STEP 1: %d\n",t);
     if (t==-1) return -1;

     //find the PUT t, EBP
     for (i=t,t=-1;i>=0;i--) {
        u_in=VG_(get_instr) (cb_in,i);
	if (u_in->opcode==PUT && u_in->val2==R_EBP) {
	    t=i;
	    break;
        }
     }
     if (t==-1) return -1;
     //find PUT t, ESP
     for (i=t,t=-1;i>=0;i--) {
        u_in=VG_(get_instr) (cb_in,i);
	if (u_in->opcode==INCEIP) return -1;
	if (u_in->opcode==PUT && u_in->val2==R_ESP) {
	    t=i;
	    break;
        }
     }

     if (t==-1) return -1;

     //get the prvious LOAD (%esp), EBP 
     for (i=t,t=-1;i>=0;i--) {
        u_in=VG_(get_instr) (cb_in,i);
	if (u_in->opcode==INCEIP) return -1;
	if (u_in->opcode==LOAD) return i;
     }
     return -1;
}	   


//find pop %ebp
static Int stack_free_space_find_index_new(UCodeBlock * cb_in, Int * esp_reg)
{
     Int i,t;
     UInstr * u_in;
     t=-1;
     Int use;
   
     //find there is a return
     for (i=VG_(get_num_instrs) (cb_in)-1;i>=0;i--) {
        u_in=VG_(get_instr) (cb_in,i);
        if (u_in->opcode==JMP ) {
	/*get wrong for __int_malloc because the JMP is not necessary ret*/
        //if (u_in->opcode==JMP && u_in->jmpkind==JmpRet) {
	   t=i;
	   break;
	}
     }
     //VG_(printf) ("STEP 1: %d\n",t);
     if (t==-1) return -1;

     //find the PUT t, EBP
     for (i=t,t=-1;i>=0;i--) {
        u_in=VG_(get_instr) (cb_in,i);
	if (u_in->opcode==PUT && u_in->val2==R_EBP) {
	    use=u_in->val1;
            t=i;
	    break;
        }
     }

     if (t==-1) return -1;
     //find the defintion for 'use', normally LD (ts), td
     for (i=t,t=-1;i>=0;i--) {
        u_in=VG_(get_instr) (cb_in,i);
	if (u_in->opcode==INCEIP) return -1;
	if (u_in->val2==use) {
	    use=u_in->val1;
            t=i;
	    break;
        }
     }

     if (t==-1) return -1;
     //find the definition for 'use', normally ESP+x, use
     //get the prvious LOAD (%esp), EBP 
     for (i=t,t=-1;i>=0;i--) {
        u_in=VG_(get_instr) (cb_in,i);
	if (u_in->opcode==INCEIP) {
            *esp_reg=use;
	    return i;
        }
	if (u_in->val2==use) return i;
     }
     return -1;
}	   

static void stack_free_space_instrument_new(UCodeBlock * cb, Int reg, fun_entry * fe)
{
     uInstr2(cb, ADD, 4, Literal, 0, TempReg, reg);
     //uInstr2(cb, ADD, 4, Literal, 0, TempReg, u_get->val1);
     uLiteral(cb, fe->alloc_lv * PER_CD_SPACE+LAST_PREDICATE);
     //uInstr2(cb, PUT, 4, TempReg, u_get->val2, ArchReg, R_ESP);
}

//To allocate space for CD, call when the function entry BB is instrumented.
static Int stack_alloc_space_find_index(UCodeBlock * cb_in, fun_entry * fe)
{
    Int i,j;
    UInstr * u_in, *tmp_in;
    Bool start_counting=False;
    Int esp_def_cnt=0;
  
    for (i=0;i<VG_(get_num_instrs)(cb_in)-1;i++) {
       u_in = VG_(get_instr)(cb_in, i);
       if (u_in->opcode!=PUT) continue;
       if (u_in->val2==R_EBP) start_counting=True;
       if (!start_counting) continue;
       tmp_in = VG_(get_instr)(cb_in, i+1);
       if (u_in->val2==R_ESP){
           if (tmp_in->opcode==STORE) 	esp_def_cnt++;
	   else {//!=STORE
	        j=i-1;
                do {
		    tmp_in=VG_(get_instr)(cb_in,j);
	            if (tmp_in->opcode==SUB && tmp_in->tag1==Literal && tmp_in->val2==u_in->val1) {//the SUB instr 	
                        fe->slack=(esp_def_cnt-1) * 4;
			fe->orig_sp_offset=tmp_in->lit32+fe->slack;
		        return j;
                    }
		    j--;
		} while (j>=0);
                sk_assert(j>=0); //always failed at this point
            }
        }
     }//for
     return -1;
}

static void stack_alloc_space_instrument(UCodeBlock * cb, UInstr * u_in, fun_entry * fe,Int t_reg)
{
                            //uInstr0(cb,INS,0);
                            uInstr2(cb, SUB, 4, Literal, 0, TempReg, u_in->val1);
                            //uLiteral(cb, 32); 
                            uLiteral(cb,fe->alloc_lv * PER_CD_SPACE+LAST_PREDICATE);
                            //uLiteral(cb,fe->alloc_lv * PER_CD_SPACE);

                            VG_(copy_UInstr)(cb, u_in);

                            //ESP should be changed before initilization becuase ESP change will allocate the page, or else you could get segment fault.        
                            //init the allocated lvs to 0
                            //uInstr0(cb,INS,0);

                            uInstr2(cb,GET, 4, ArchReg, R_EBP, TempReg, t_reg);

			    VG_(ccall_RL_0) (cb, (Addr) & XZ_HELPER_init_lvs, t_reg, (UInt)fe, 2);

}


//it is called when ESP is the def port
static void stack_free_space_instrument(UCodeBlock * cb_in, UCodeBlock * cb, fun_entry * fe ,Int idx, UInstr * u_in)
{
    UInstr * tmp_in,*tmp_in1;
    Int j;

                   j=idx-1;
                   do {
                       tmp_in = VG_(get_instr)(cb_in, j);
                       tmp_in1 = VG_(get_instr)(cb_in, j-1);
                       if (tmp_in1->opcode==GET && tmp_in1->val1==R_ESP && tmp_in->opcode==ADD && tmp_in->tag1==Literal && tmp_in->val2==u_in->val1 && tmp_in->lit32==fe->orig_sp_offset-fe->slack ) {//so ugly hacking!!
                            //lv_finish=True;
                            //it is the substraction of the sp
                            uInstr0(cb,INS,0);
                            uInstr2(cb, ADD, 4, Literal, 0, TempReg, u_in->val1);
                            //uLiteral(cb, 32); 
                            uLiteral(cb,fe->alloc_lv * PER_CD_SPACE+LAST_PREDICATE);
                            return;
                       }
                       j--;
                    } while (j>=1);

                    //sk_assert (j>=1);
}
#ifdef TRACK_ARGV
static void instrument_main_entry(UCodeBlock * cb, Int  r_tmp, Int r_tmp1)
{
	  //VG_(printf)("INSTRUMENT MAIN\n");
          uInstr2(cb, GET, 4, ArchReg, R_ESP, TempReg, r_tmp);
          uInstr2(cb, LEA1, 4, TempReg, r_tmp, TempReg, r_tmp1);
          uLiteral(cb, 0x4);
          uInstr2(cb, LOAD, 4, TempReg, r_tmp1, TempReg, r_tmp1);
          uInstr2(cb, LEA1, 4, TempReg, r_tmp, TempReg, r_tmp);
          uLiteral(cb,0x8);
          uInstr2(cb, LOAD, 4, TempReg, r_tmp, TempReg, r_tmp);
          VG_(ccall_RR_0) (cb, (Addr) & XZ_HELPER_track_argv, r_tmp1, r_tmp,2);
}
#endif 
 


static void instrument_predicate_with_ii(UCodeBlock * cb, UCodeBlock * cb_in, instr_info * bb, instr_info * ii, Int * r_tmp, Int * r_tmp1,Int * r_tmp2) 
{
          /*sk_assert(bb->lv_idx!=LV_NOT_ASSIGNED);*/      
          uInstr2(cb, GET, 4, ArchReg, R_EBP, TempReg, *r_tmp); 
	  //DEBUG
//#define DEBUG
//#undef DEBUG
#ifdef DEBUG
          VG_(ccall_RL_0) (cb, (Addr) & XZ_HELPER_debug_EBP_get_value, *r_tmp,0, 1);
#endif
          /*store instr_info into stack*/ 
          uInstr2(cb, LEA1, 4, TempReg, *r_tmp, TempReg, *r_tmp1);
          uLiteral(cb, (PER_CD_SPACE*bb->lv_idx+LAST_PREDICATE+CD_II_OFFSET-4));
	  uInstr2(cb, MOV, 4, Literal, 0, TempReg, * r_tmp2);
          uLiteral(cb, ii);
          uInstr2(cb, STORE, 4, TempReg,*r_tmp2, TempReg, *r_tmp1);
          //DEBUG
#ifdef DEBUG
          //VG_(ccall_LRR_0)(cb, (Addr) & XZ_HELPER_debug_store_value, 0, *r_tmp1, ii,3);
#endif
          //store instance into stack
	  uInstr2(cb, LEA1, 4, TempReg, *r_tmp, TempReg, *r_tmp1);
          uLiteral(cb, (PER_CD_SPACE*bb->lv_idx+LAST_PREDICATE+CD_INST_OFFSET-4));

          VG_(ccall_L_R) (cb, (Addr) & XZ_HELPER_get_inst, ii, *r_tmp2, 1);
          uInstr2(cb, STORE, 4, TempReg, *r_tmp2, TempReg, *r_tmp1);
          //DEBUG
#ifdef DEBUG
          //VG_(ccall_LRR_0)(cb, (Addr) & XZ_HELPER_debug_store_value, 0, *r_tmp1, *r_tmp2,3);
#endif
          //store timestamp into stack
	  uInstr2(cb, LEA1, 4, TempReg, *r_tmp, TempReg, *r_tmp1);
          uLiteral(cb, (PER_CD_SPACE*bb->lv_idx+LAST_PREDICATE+CD_TIMESTAMP_OFFSET-4));
          VG_(ccall_R_R) (cb, (Addr) &XZ_HELPER_get_timestamp, *r_tmp, *r_tmp,1);
          uInstr2(cb, STORE, 4, TempReg, *r_tmp, TempReg, *r_tmp1);
          //DEBUG
#ifdef DEBUG
          //VG_(ccall_LRR_0)(cb, (Addr) & XZ_HELPER_debug_store_value, 0, *r_tmp1, *r_tmp ,3);
#endif
#undef DEBUG
}




static void instrument_predicate (UCodeBlock * cb, UCodeBlock * cb_in, instr_info * bb, UInstr * u_in, Int * r_tmp, Int * r_tmp1) 
{
          Int r_dest;
 
          /*find the destination reg*/ 
	  if (u_in->tag3==TempReg) { 
	      r_dest=u_in->val3; 
	  } else if (u_in->tag2==TempReg) {
	      r_dest=u_in->val2;
	  } else if (u_in->tag1==TempReg) {
	      r_dest=u_in->val1;
          } else {
	      //print_cb(cb_in);
	      //VG_(skin_panic) ("Could not find the definition port.");
	      return;
          } 
          /*sk_assert(bb->lv_idx!=LV_NOT_ASSIGNED);*/      
          uInstr2(cb, GET, 4, ArchReg, R_EBP, TempReg, *r_tmp); 
	  //DEBUG
//#define DEBUG
//#undef DEBUG
#ifdef DEBUG
          VG_(ccall_RL_0) (cb, (Addr) & XZ_HELPER_debug_EBP_get_value, *r_tmp,0, 1);
#endif
          /*store instr_info into stack*/ 
          uInstr2(cb, LEA1, 4, TempReg, *r_tmp, TempReg, *r_tmp1);
          uLiteral(cb, (PER_CD_SPACE*bb->lv_idx+LAST_PREDICATE+CD_II_OFFSET-4));
          uInstr2(cb, STORE, 4, TempReg, SHADOW(r_dest), TempReg, *r_tmp1);
          //DEBUG
#ifdef DEBUG
          //VG_(ccall_LRR_0)(cb, (Addr) & XZ_HELPER_debug_store_value, 0, *r_tmp1, SHADOW(r_dest),3);
#endif
          //store instance into stack
	  uInstr2(cb, LEA1, 4, TempReg, *r_tmp, TempReg, *r_tmp1);
          uLiteral(cb, (PER_CD_SPACE*bb->lv_idx+LAST_PREDICATE+CD_INST_OFFSET-4));
          uInstr2(cb, STORE, 4, TempReg, SHADOW1(r_dest), TempReg, *r_tmp1);
          //DEBUG
#ifdef DEBUG
          //VG_(ccall_LRR_0)(cb, (Addr) & XZ_HELPER_debug_store_value, 0, *r_tmp1, SHADOW1(r_dest),3);
#endif
          //store timestamp into stack
	  uInstr2(cb, LEA1, 4, TempReg, *r_tmp, TempReg, *r_tmp1);
          uLiteral(cb, (PER_CD_SPACE*bb->lv_idx+LAST_PREDICATE+CD_TIMESTAMP_OFFSET-4));
          VG_(ccall_R_R) (cb, (Addr) &XZ_HELPER_get_timestamp, *r_tmp, *r_tmp,1);
          uInstr2(cb, STORE, 4, TempReg, *r_tmp, TempReg, *r_tmp1);
          //DEBUG
#ifdef DEBUG
          //VG_(ccall_LRR_0)(cb, (Addr) & XZ_HELPER_debug_store_value, 0, *r_tmp1, *r_tmp ,3);
#endif
#undef DEBUG
}


/* Multiple ucode could correspond to one x86 code*/
#define TRACK_INSTR \
	consider_defined=True; \
	if (!ii) { \
            get_bb_info(cur_x86_addr,&cur_fileCC, &cur_fnCC, &line); \
	    ii=get_instr_info(cur_x86_addr,fe,line,INVALID_UPORT,&is_new_instr,fe->cd_available); \
  	    /*VG_(printf) ("ii=%x, ii->bb=%x, bb=%x\n",ii->orig_addr, ii->bb, bb);*/ \
	    if (fe->cd_available && ii->bb!=bb) { \
		bb=ii->bb; \
                predicate_idx=search_for_predicate(cb_in,i); \
	    } \
        }

#define TRACK_INSTR_FORCE_MODE \
	consider_defined=True; \
	if (!ii) { \
            get_bb_info(cur_x86_addr,&cur_fileCC, &cur_fnCC, &line); \
	    ii=get_instr_info(cur_x86_addr,fe,line,INVALID_UPORT,&is_new_instr,fe->cd_available); \
        }

#define UPDATE_INSTR_UPORT \
        if (ii && ii->use_port==INVALID_UPORT) { \
	   ii->use_port=cur_uport; \
	}
  

// DENNIS modified to do nothing, since we're not recording control dependences here
#define SETUP_CD  ; /* \
         if (fe->cd_available && i>stack_init_idx && !stack_freed) { \
            uInstr2(cb, GET, 4, ArchReg, R_EBP, TempReg, r_tmp); \
	    VG_(ccall_LR_0) (cb, (Addr)&XZ_HELPER_setup_CD_new, (UInt)ii, r_tmp,2); \
         } */

#define INCREMENT_COUNTER \
	    VG_(ccall_L_0) (cb, (Addr) & XZ_HELPER_increment_inst, (UInt) ii, 1);
	    
//MOVE the shadow of rs to rd depends on tag
#define INSTRUMENT_OPERAND_SHADOW(tag, rs, rd) \
     switch (tag) {  \
	case TempReg: \
	    uInstr2(cb, MOV, 4, TempReg, SHADOW(rs),TempReg, SHADOW(rd)); \
	    uInstr2(cb, MOV, 4, TempReg, SHADOW1(rs), TempReg, SHADOW1(rd)); \
	    break; \
        case Literal: \
	    TRACK_INSTR \
            XZ_set_instr_inst(cb,ii,SHADOW(rd), SHADOW1(rd)); \
	    break; \
	case ArchReg: \
	    uInstr3(cb, GETV, 4, ArchReg, rs, TempReg, SHADOW(rd), TempReg, SHADOW1(rd)); \
	    break; \
      }
      

#define INSTRUMENT_OPERAND_SHADOW_FORCE_MODE(tag, rs, rd) \
     switch (tag) {  \
	case TempReg: \
	    uInstr2(cb, MOV, 4, TempReg, SHADOW(rs),TempReg, SHADOW(rd)); \
	    uInstr2(cb, MOV, 4, TempReg, SHADOW1(rs), TempReg, SHADOW1(rd)); \
	    break; \
        case Literal: \
	    TRACK_INSTR_FORCE_MODE \
            XZ_set_instr_inst(cb,ii,SHADOW(rd), SHADOW1(rd)); \
	    break; \
	case ArchReg: \
	    uInstr3(cb, GETV, 4, ArchReg, rs, TempReg, SHADOW(rd), TempReg, SHADOW1(rd)); \
	    break; \
      }


#define INSTRUMENT_DEBUG_TERMINATE \
      VG_(ccall_0_0) (cb, (Addr) & XZ_HELPER_debug_terminate);

#define INSTRUMENT_DEBUG_GET\
      VG_(ccall_LL_0) (cb, (Addr) & XZ_HELPER_debug_get, i, u_in->val1,2);

#define INSTRUMENT_DEBUG_PUT\
      VG_(ccall_LRR_0) (cb, (Addr) & XZ_HELPER_debug_put, i, SHADOW(u_in->val1),SHADOW1(u_in->val1), 3);

#define INSTRUMENT_DEBUG \
      VG_(ccall_L_0) (cb, (Addr) & XZ_HELPER_debug, i, 1);




// record a data dependence on the specified register
#define REG_DEP(use_reg) \
    internal_defined=False;  \
    /*VG_(printf) ("XXXXXXXXX %d CUR_UPORT=%d\n",i, cur_uport);*/ \
    for (k=0;k<cur_def;k++) {  \
	if (use_reg == internal_defs[k]) {  \
             internal_defined=True;  \
	     break;             \
        } \
    } \
    if (!internal_defined) {  /*defined by previous x86 instruction*/ \
	      /*VG_(printf) ("\tXXXXXXXXX %d CUR_UPORT=%d\n",i, cur_uport);*/ \
	      /* DENNIS added this chunk of code to save the archReg that is being used here */ \
	      Int archRegFound = ii->usedRegister[cur_uport]; /* could be LOCATION_FLAG or LOCATION_UNDEFINED */ \
	      Int dj_i; \
		   for(dj_i = 0; dj_i < 7; dj_i++) \
		   { \
		      if(tempRegUse[dj_i] == use_reg) \
		      { \
		         archRegFound = archRegUse[dj_i]; \
		         break; \
		      } \
		   } \
		   if(archRegFound == LOCATION_UNDEFINED) \
		   { \
	          VG_(printf)("Error: could not find arch reg associated with used temp reg t%d! 0x%x\n", use_reg, ii->orig_addr); \
	          VG_(exit)(0); \
		   } \
	      ii->usedRegister[cur_uport] = archRegFound; \
	      /* record the use if we're NOT in force mapping mode, or if we're in force-mapping mode AND in the main prog file */ \
	      if((valgrindMode != MODE_FORCE_MAPPING) || (valgrindMode == MODE_FORCE_MAPPING && insideMainProgFile == True)) \
	      { \
	         VG_(ccall_LRR_0) (cb, helper_reg_dep_ports[cur_uport++], (UInt)ii, SHADOW(use_reg), SHADOW1(use_reg),3); \
	      } \
    }


#define REG_DEP_FORCE_MODE(use_reg) \
    internal_defined=False;  \
    /*VG_(printf) ("XXXXXXXXX %d CUR_UPORT=%d\n",i, cur_uport);*/ \
    for (k=0;k<cur_def;k++) {  \
	if (use_reg == internal_defs[k]) {  \
             internal_defined=True;  \
	     break;             \
        } \
    } \
    if (!internal_defined) {  /*defined by previous x86 instruction*/ \
	      /*VG_(printf) ("\tXXXXXXXXX %d CUR_UPORT=%d\n",i, cur_uport);*/ \
	      /* DENNIS added this chunk of code to save the archReg that is being used here */ \
	      Int archRegFound = ii->usedRegister[cur_uport]; /* could be LOCATION_FLAG or LOCATION_UNDEFINED */ \
	      Int dj_i; \
		   for(dj_i = 0; dj_i < 7; dj_i++) \
		   { \
		      if(tempRegUse[dj_i] == use_reg) \
		      { \
		         archRegFound = archRegUse[dj_i]; \
		         break; \
		      } \
		   } \
		   if(archRegFound == LOCATION_UNDEFINED) \
		   { \
	          VG_(printf)("Error: could not find arch reg associated with used temp reg t%d! 0x%x\n", use_reg, ii->orig_addr); \
	          VG_(exit)(0); \
		   } \
	      ii->usedRegister[cur_uport] = archRegFound; \
	         VG_(ccall_LRR_0) (cb, helper_reg_dep_ports[cur_uport++], (UInt)ii, SHADOW(use_reg), SHADOW1(use_reg),3); \
    } \



#define SETUP_TWO_EDGES \
	    REG_DEP(u_in->val1) \
	    REG_DEP(u_in->val2) \
	    XZ_set_instr_inst(cb,ii,SHADOW(u_in->val2), SHADOW1(u_in->val2)); 



// record a data dependence on the specified memory location
#define MEM_DEP(addr_reg) \
      /* DENNIS added this chunk of code to save the archReg that is being used here */ \
      /* note: the archReg value is NEGATED to indicate that the use is from memory, not the register itself */ \
	      Int archRegFound = LOCATION_UNDEFINED; \
	      Int dj_i; \
		   for(dj_i = 0; dj_i < 7; dj_i++) \
		   { \
		      if(tempRegUse[dj_i] == addr_reg) \
		      { \
		         archRegFound = archRegUse[dj_i]; \
		         break; \
		      } \
		   } \
		   if(archRegFound == LOCATION_UNDEFINED) \
		   { \
	          VG_(printf)("Error: memory use: could not find arch reg associated with used temp reg t%d! 0x%x\n", addr_reg, ii->orig_addr); \
	          VG_(exit)(0); \
		   } \
		   /* make sure to make the archReg *negative* so we know it refers to a memory location */ \
		   if(archRegFound >= 0) \
		   { \
		      archRegFound = -archRegFound; \
		      /* special check for R_EAX, since this has value 0 and thus cannot be negated */ \
		      if(archRegFound == R_EAX) \
		      {\
		         archRegFound = R_EAX_MEM; \
		      }\
		   } \
	      ii->usedRegister[cur_uport] = archRegFound; \
	      /* record the use if we're NOT in force mapping mode, or if we're in force-mapping mode AND in the main prog file */ \
	      if((valgrindMode != MODE_FORCE_MAPPING) || (valgrindMode == MODE_FORCE_MAPPING && insideMainProgFile == True)) \
	      { \
	         VG_(ccall_LLR_0) (cb, (Addr) &XZ_HELPER_load_mem_dep, (UInt)ii, cur_uport++, addr_reg, 3); \
	      } \



// record a data dependence on the specified memory location
#define MEM_DEP_FORCE_MODE(addr_reg) \
      /* DENNIS added this chunk of code to save the archReg that is being used here */ \
      /* note: the archReg value is NEGATED to indicate that the use is from memory, not the register itself */ \
	      Int archRegFound = LOCATION_UNDEFINED; \
	      Int dj_i; \
		   for(dj_i = 0; dj_i < 7; dj_i++) \
		   { \
		      if(tempRegUse[dj_i] == addr_reg) \
		      { \
		         archRegFound = archRegUse[dj_i]; \
		         break; \
		      } \
		   } \
		   if(archRegFound == LOCATION_UNDEFINED) \
		   { \
	          VG_(printf)("Error: memory use: could not find arch reg associated with used temp reg t%d! 0x%x\n", addr_reg, ii->orig_addr); \
	          VG_(exit)(0); \
		   } \
		   /* make sure to make the archReg *negative* so we know it refers to a memory location */ \
		   if(archRegFound >= 0) \
		   { \
		      archRegFound = -archRegFound; \
		      /* special check for R_EAX, since this has value 0 and thus cannot be negated */ \
		      if(archRegFound == R_EAX) \
		      {\
		         archRegFound = R_EAX_MEM; \
		      }\
		   } \
	      ii->usedRegister[cur_uport] = archRegFound; \
	      VG_(ccall_LLR_0) (cb, (Addr) &XZ_HELPER_load_mem_dep, (UInt)ii, cur_uport++, addr_reg, 3); \




/*
#define DEFINED_BY_EBP(reg) \
          ebp_used=False; \
	  reg_defed=False; \
	  for (j=i-1;j>=0;j--) { \
	      tmp_in=VG_(get_instr)(cb_in,j); \
              k=VG_(get_reg_usage) ( tmp_in, TempReg, instr_regs, instr_regs_iswrite); \
	      for (l=0;l<k; l++) { \
		 if (instr_regs_iswrite[l] && instr_regs[l]==reg) { \
		     reg_defed=True; \
                 } \
              }\
	      if (reg_defed) { \
		  if (tmp_in->val1==R_EBP || tmp_in->val2==R_EBP || tmp_in->val3==R_EBP) \
			ebp_used=True; \
		  break; \
              } \
          }
      
        
		   
#define OFFSET_EBP_ACCESS_LEA1 \
	  if (fe->cd_available) \
		DEFINED_BY_EBP(u_in->val1) \
          if (ebp_used  && ((Int) u_in->lit32)>0) { \
		uLiteral(cb, ((Int)u_in->lit32)+fe->alloc_lv*PER_CD_SPACE+LAST_PREDICATE); \
          }

#define OFFSET_EBP_ACCESS_LEA2 \
	  if (fe->cd_available) { \
		DEFINED_BY_EBP(u_in->val1) \
	        if (!ebp_used)  \
		    DEFINED_BY_EBP(u_in->val2) \
          } \
          if (ebp_used  && ((Int) u_in->lit32)>0) { \
		uLiteral(cb, ((Int)u_in->lit32)+fe->alloc_lv*PER_CD_SPACE+LAST_PREDICATE); \
          }
*/

#define OFFSET_EBP_ACCESS_LEA1 \
	  /* After the function entry BB*/ \
        /*	VG_(printf)("TRY OFFSET LEA1 %x ebp_reg=%d\n",cur_x86_addr,ebp_reg);*/ \ 
	  if (ebp_reg==u_in->val1  && ((Int) u_in->lit32)>0) { \
		/*VG_(printf)("OFFSET LEA1 %x\n",cur_x86_addr);*/ \
		uLiteral(cb, ((Int)u_in->lit32)+fe->alloc_lv*PER_CD_SPACE+LAST_PREDICATE); \
          }
	
#define OFFSET_EBP_ACCESS_LEA2 \
	  /* After the function entry BB*/ \
	  if ((ebp_reg==u_in->val1 || ebp_reg==u_in->val2) && ((Int) u_in->lit32)>0) { \
		/*VG_(printf)("OFFSET LEA2\n");*/ \
		uLiteral(cb, ((Int)u_in->lit32)+fe->alloc_lv*PER_CD_SPACE+LAST_PREDICATE); \
          }
	
#define MAX_INTERNAL_DEF 20

// DENNIS added this helper function so that during instrumentation, if a given
// dependence is about to be recorded, we first call this function to determine
// whether the x86 instruction corresponding to the specified ucode instruction is
// one that we're ignoring or not.  If so, we do *not* output the dependence, so that
// this information does not propagate to the value profile data
Bool ignoreThisX86Instruction(UInstr *instr)
{
   if(instr->x86Instr == x86_PUSH)
   {
      // DON'T IGNORE IT!
      //return True;
      return False;
   }
   else
   {
      return False;
   }
}

// DENNIS helper function called from SK(instrument)
#define DJ_RESET_ARCH_REGS \
   for(dj_i=0; dj_i < 7; dj_i++) archRegUse[dj_i] = LOCATION_UNDEFINED; \
   for(dj_i=0; dj_i < 7; dj_i++) tempRegUse[dj_i] = LOCATION_UNDEFINED; \
   for(dj_i=0; dj_i < 7; dj_i++) archRegDef[dj_i] = LOCATION_UNDEFINED; \
   for(dj_i=0; dj_i < 7; dj_i++) tempRegDef[dj_i] = LOCATION_UNDEFINED; \
   nextUsePort = 0; \
   nextDefPort = 0;



// In the main instrumentation function for def-use pairs, this helper function is called to
// record the defined value (and flags if necessary) when we get the INCEIP opcode
void recordDefinition(instr_info * ii, Int last_define_reg, Int last_define_val, Int tempRegDef[],
                      Int archRegDef[], UCodeBlock* cb, Int last_def_size, Bool whetherFlagsWritten,
                      Bool definedLocationIsFPUStack)
{
   int dj_i;

   if (ii && last_define_reg != -1)
	{
	   // last thing defined is a REGISTER
	        
	   if(last_define_val != -1)
	   {
	      VG_(printf)("Dennis error: last_define_reg is not -1, but so is last_define_val!\n");
	      VG_(exit)(0);
	   }
	    
		//VG_(printf) ("TRACE VALUE @ %d, size=%d\n",last_define, last_def_size);
		     
		// find the last defined archReg corresponding to the last defined temp reg
		Int archRegFound = LOCATION_FLAG;
		for(dj_i = 0; dj_i < 7; dj_i++)
		{
         if(tempRegDef[dj_i] == last_define_reg)
		   {
		      archRegFound = archRegDef[dj_i];
		      break;
		   }
		}
		if(archRegFound == LOCATION_FLAG)
		{
	      // This can occur when the last temp register defined is actually NOT written back to an
	      // archReg, which occurs for CMP, TEST, AND, etc.  In this case, *flags* are written so
	      // that a jump can subsequently occur.
		}
		
		if(definedLocationIsFPUStack == True)
		{
			// special case: ensure the defined location is type LOCATION_FPU
			archRegFound = LOCATION_FPU;
		}
		
		ii->definedRegister = archRegFound;
		
	    VG_(ccall_LR_0) (cb, helper_trace_values[last_def_size], (UInt)ii, last_define_reg, 2);
	        
	   // record the defined flags, if any were defined
	   if(whetherFlagsWritten == True)
	   {
         // get the defined flags
         Int flagTempReg = 1; /* register t1 is not used elsewhere, so I'll use it here */
         uInstr1(cb, GETF, 4, TempReg, flagTempReg);
         uFlagsRWU(cb, FlagsALL, FlagsEmpty, FlagsEmpty); /* specify the previous GETF instruction should READ ALL FLAGS */	            
	      VG_(ccall_LR_0) (cb, DJ_set_flags_defined, (UInt)ii, flagTempReg,2);
	   }
   } // end if the last thing defined was a register
   else if(ii && last_define_val != -1)
   {
      // last thing defined is MEMORY
           
      if(last_define_reg != -1)
	   {
	      VG_(printf)("Dennis error: last_define_val is not -1, but so is last_define_reg!\n");
	      exit(0);
	   }
	   
	   ii->definedRegister = LOCATION_MEM;

	   VG_(ccall_LR_0) (cb, helper_trace_values[last_def_size], (UInt)ii, last_define_val, 2);
	        
	   // record the defined flags, if any were defined
	   if(whetherFlagsWritten == True)
	   {
	      // get the defined flags
         Int flagTempReg = 1; /* register t1 is not used elsewhere, so I'll use it here */
         uInstr1(cb, GETF, 4, TempReg, flagTempReg);
         uFlagsRWU(cb, FlagsALL, FlagsEmpty, FlagsEmpty); /* specify the previous GETF instruction should READ ALL FLAGS */	            
	      VG_(ccall_LR_0) (cb, DJ_set_flags_defined, (UInt)ii, flagTempReg,2);
	   }
   } // end if the last thing defined was memory
}

void myAssert(Bool condition, char *message)
{
   if(condition == False)
   {
      //VG_(message)(Vg_UserMsg,"Error: myAssert failed: %s\n", message);
      VG_(printf)("Error: myAssert failed: %s\n", message);
      VG_(exit)(0);
   }
}

UCodeBlock* DJ_instrument_suppress(UCodeBlock* cb_in, Addr orig_addr)
{
   Int i;
   UCodeBlock *cb;
   Addr curr_x86_addr;
   Addr prev_x86_addr;
   UInstr* u_in;
   x86InstrStruct * x86InstrStructNode;
   
   cb = VG_(setup_UCodeBlock)(cb_in);
   
   prev_x86_addr = 0x0;
   curr_x86_addr=orig_addr;

   Int line;
   fileCC * cur_fileCC;
   fnCC * cur_fnCC;
   
   Int lastLine = -1;

   for (i = 0; i < VG_(get_num_instrs)(cb_in); i++)
   {
      if(prev_x86_addr != curr_x86_addr)
      {
         prev_x86_addr = curr_x86_addr;
         get_bb_info(curr_x86_addr,&cur_fileCC, &cur_fnCC, &line);
         setX86InstrInfo(curr_x86_addr, cur_fileCC->file, cur_fnCC->fn, line);
    
         /*
         if(curr_x86_addr == 0x810231d)
         {
            VG_(printf)("Valgrind here!!!! %s %s %d\n", cur_fileCC->file, cur_fnCC->fn, line);
         }
         */
         
         // If we're at a new stmt, record it in the output file
         //if(VG_(strcmp)(cur_fileCC->file, "mysqld.cc") == 0 /*&& line == 1419*/)
         //if(VG_(strcmp)(cur_fileCC->file, "main.c") == 0)
         //if(VG_(strcmp)(cur_fileCC->file, "axel.c") == 0)
         //{
         //   FILE *tempFile = fopen("#vg_start_record_stmts.txt", "w");
         //   fclose(tempFile);
            //VG_(printf)("%d %d\n",VG_(client_base),VG_(client_end));
            //VALGRIND_DISCARD_TRANSLATIONS(VG_(client_base), VG_(client_end) - VG_(client_base));
         //}

         if(VG_(strcmp)(cur_fileCC->file, "???") != 0 && line != lastLine)
         {
            // instrument here to record the current statement information when executing.
            VG_(ccall_LLL_0) (cb, (Addr) & DJ_record_stmt_info, cur_fileCC->file, cur_fnCC->fn, line, 3);
            
            //FILE *filePointer = fopen("#vg_stmt_trace.txt", "a");
            //fprintf(filePointer, "%s/%s/%d\n", cur_fileCC->file, cur_fnCC->fn, line);
            //fclose(filePointer);
            
            /*
            if(VG_(strcmp)(cur_fnCC->fn, "next_event") == 0)
            {
               VG_(printf)("%s/%s/%d\n", cur_fileCC->file, cur_fnCC->fn, line);
            }
            */
            
            //VG_(printf)("%s/%s/%d\n", cur_fileCC->file, cur_fnCC->fn, line);
            

            
            lastLine = line;
         }
      }
   
      u_in = VG_(get_instr)(cb_in, i);
      
      //VG_(ccall_LL_0) (cb, (Addr) & DJ_debug_trace, curr_x86_addr, u_in->opcode, 2);
      
      
      /* UCODE instructions I don't think I need to specially handle for now (may discover later that I'll need to handle them)
         NOP, LOCK, GETF, PUTF, FPU, FPU_R, FPU_W, all MMX instrs, all SSE instrs,
         JIFZ, CALLM_S, CALLM_E, CLEAR, CALLM, CCALL <CONTIUE HERE!!!!>
      */
      switch(u_in->opcode)
      {
         case GET: // DONE! *********************************************************
         {
            // moves from an archReg to a tempReg
            myAssert(u_in->tag1 == ArchReg, "GET tag1 is not ArchReg when should be!");
            myAssert(u_in->tag2 == TempReg, "GET tag2 is not TempReg when should be!");
            
            // just transfer the suppression mark from source to destination
            VG_(ccall_LL_0) (cb, (Addr) & DJ_transfer_mark_arch_to_temp, u_in->val1, u_in->val2, 2);
            
            VG_(copy_UInstr)(cb, u_in);
            break;
         } // end case GET
         case PUT: // DONE! *********************************************************
         {
            // moves from a tempReg to an archReg
            myAssert(u_in->tag1 == TempReg, "PUT tag1 is not TempReg when should be!");
            myAssert(u_in->tag2 == ArchReg, "PUT tag2 is not ArchReg when should be!");
            
            // just transfer the suppression mark from source to destination
            VG_(ccall_LL_0) (cb, (Addr) & DJ_transfer_mark_temp_to_arch, u_in->val1, u_in->val2, 2);
            
            VG_(copy_UInstr)(cb, u_in);
            break;
         } // end case PUT
         case LOAD: // DONE! *********************************************************
         {
            // LOAD reg(address loaded from), reg(destination to send value)
            myAssert(u_in->tag1 == TempReg, "LOAD tag1 is not TempReg when should be!");
            myAssert(u_in->tag2 == TempReg, "LOAD tag2 is not TempReg when should be!");
            
            // add special instrumentation here to set the current tempReg that is being
            // written to, and to set the current x86 instruction
            VG_(ccall_LL_0) (cb, (Addr) & DJ_set_dest_reg_for_load, u_in->val2, curr_x86_addr, 2);
            
            // see if the temp register or the address contained in it are suppressed.  If
            // so, then load from the scratch location and mark the destination tempReg as
            // suppressed.  If not, then perform the original load and mark the destination
            // tempReg as not suppressed.  Call a helper function that does this, and returns
            // the address we should actually load from.
            Int returnReg = 1; // t1 won't be used elsewhere here, so we make use of it
            VG_(ccall_LR_R) (cb, (Addr) & DJ_handle_load_to_possibly_suppress, u_in->val1, u_in->val1, returnReg, 2);
            
            // perform the load (either from the existing location address, or from a different (scratch) location)
            uInstr2(cb, LOAD, u_in->size, TempReg, returnReg, TempReg, u_in->val2);
            
            
            // For forcing a function call return value to jump to a known valid location.
            if(curr_x86_addr == 0x8053976) // address of the ret instruction
            {
               uInstr2(cb, MOV, u_in->size, Literal, (Int)(0), TempReg, u_in->val2);
               uLiteral(cb, (UInt)(0x80548b9)); // the known valid location to jump back to
            }
            
            
            VG_(ccall_R_0) (cb, (Addr) & DJ_print_value, u_in->val2, 1);
            
            break;
         } // end case LOAD
         case STORE: // DONE! *********************************************************
         {
            // STORE reg(source containing value), reg(destination address to store to)
            myAssert(u_in->tag1 == TempReg, "STORE tag1 is not TempReg when should be!");
            myAssert(u_in->tag2 == TempReg, "STORE tag2 is not TempReg when should be!");
            
            // add special instrumentation here to set a flag in case the TempReg being
            // read from, or the tempReg containing the destination address, is marked as suppressed
            VG_(ccall_LL_0) (cb, (Addr) & DJ_set_flag_for_store_if_suppressed, u_in->val1, u_in->val2, 2);
            
            // add instrumentation that will handle the possibility of this being an
            // initial suppression point, and it will return the appropriate 
            // address to store to.  If this is not an initial suppression point, it will
            // also output information about the stored value.
            Int returnReg = 1; // t1 won't be used elsewhere here, so we make use of it
            VG_(ccall_LR_R) (cb, (Addr) & DJ_handle_initial_suppression_point, curr_x86_addr, u_in->val2, returnReg, 2);
            
            // perform the store (either to the existing location address, or to a different (scratch) location)
            uInstr2(cb, STORE, u_in->size, TempReg, u_in->val1, TempReg, returnReg);
            
            VG_(ccall_R_0) (cb, (Addr) & DJ_print_value, u_in->val1, 1);
            
            break;
         } // end case STORE
         case MOV: // DONE! *********************************************************
         {
            // move constant or tempReg into a tempReg
            myAssert(u_in->tag1 == TempReg || u_in->tag1 == Literal, "MOV tag1 is not TempReg/Literal when should be!");
            myAssert(u_in->tag2 == TempReg, "MOV tag2 is not TempReg when should be!");
            
            if(u_in->tag1 == TempReg)
            {
               // just transfer the suppression mark from source to destination
               VG_(ccall_LL_0) (cb, (Addr) & DJ_transfer_mark_temp_to_temp, u_in->val1, u_in->val2, 2);
            }
            else if(u_in->tag1 == Literal)
            {
               // just mark the destination temp as not suppressed
               VG_(ccall_L_0) (cb, (Addr) & DJ_mark_not_suppressed_temp, u_in->val2, 1);
            }
            
            VG_(copy_UInstr)(cb, u_in);
            break;
         } // end case MOV
         case CMOV: // DONE! *********************************************************
         {
            // conditionally move from one tempReg to another
            myAssert(u_in->tag1 == TempReg, "CMOV tag1 is not TempReg when should be!");
            myAssert(u_in->tag2 == TempReg, "CMOV tag2 is not TempReg when should be!");
            
            // just transfer the suppression mark from source to destination
            VG_(ccall_LL_0) (cb, (Addr) & DJ_transfer_mark_temp_to_temp, u_in->val1, u_in->val2, 2);
            
            VG_(copy_UInstr)(cb, u_in);
            break;
         } // end case CMOV
         case MUL: // DONE! *********************************************************
         case UMUL:
         case ADD:
         case ADC:
         case SUB:
         case SBB:
         case AND:
         case OR:
         case XOR:
         case SHL:
         case SHR:
         case SAR:
         case ROL:
         case ROR:                           
         case RCL:
         case RCR:
         {
            // (tempReg, archReg, or Literal) OP tempReg2 = tempReg2
            myAssert(u_in->tag1 == TempReg || u_in->tag1 == ArchReg || u_in->tag1 == Literal, "MUL... tag1 is not TempReg/ArchReg/Literal when should be!");
            myAssert(u_in->tag2 == TempReg, "MUL... tag2 is not TempReg when should be!");
            
            if(u_in->tag1 == TempReg)
            {
               // just transfer the suppression mark from sources to destination
               VG_(ccall_LLL_0) (cb, (Addr) & DJ_transfer_mark_temp_temp_to_temp, u_in->val1, u_in->val2, u_in->val2, 3);
            }
            else if(u_in->tag1 == ArchReg)
            {
               // just transfer the suppression mark from sources to destination
               VG_(ccall_LLL_0) (cb, (Addr) & DJ_transfer_mark_arch_temp_to_temp, u_in->val1, u_in->val2, u_in->val2, 3);
            }
            else if(u_in->tag1 == Literal)
            {
               // just transfer the suppression mark from source to destination
               VG_(ccall_LL_0) (cb, (Addr) & DJ_transfer_mark_temp_to_temp, u_in->val2, u_in->val2, 2);
            }
            
            VG_(copy_UInstr)(cb, u_in);
            break;
         } // end case MUL, UMUL, ADD, ADC, etc...
         case NOT: // DONE! *********************************************************
         case NEG:
         case INC:
         case DEC:
         case CC2VAL:
         case WIDEN:
         case PUSH:
         case POP:
         {
            // Doesn't need to be specially handled, since only has a single source/dest temp reg.
            // So, the temp reg will already be marked appropriately as suppressed or not.
            myAssert(u_in->tag1 == TempReg, "NOT tag1 is not TempReg when should be!");
            
            VG_(copy_UInstr)(cb, u_in);
            break;
         } // end case NOT, NEG, INC, DEC
         case BSWAP: // DONE! *********************************************************
         {
            //VG_(printf)("NEED TO HANDLE CASE BSWAP!!!\n");
            //VG_(exit)(0);
         }
         case JMP: // DONE! *********************************************************
         {
            // no need to do anything related to suppression here
            
            // increment instance for curr x86 instruction
            x86InstrStructNode = getX86InstrInfo(curr_x86_addr);
            myAssert(x86InstrStructNode != NULL, "in case JMP, could not find x86InstrStruct in hashtable!");
            VG_(ccall_L_0) (cb, (Addr) & XZ_HELPER_increment_inst, (UInt)(x86InstrStructNode), 1);
            
            VG_(copy_UInstr)(cb, u_in);
            break;
         } // end case JMP
         case LEA1: // DONE! *********************************************************
         {
            // const + TempReg1 = TempReg2
            
            myAssert(u_in->tag1 == TempReg, "LEA1 tag1 is not TempReg when should be!");
            myAssert(u_in->tag2 == TempReg, "LEA1 tag2 is not TempReg when should be!");
            
            // just transfer the suppression mark from source to destination
            VG_(ccall_LL_0) (cb, (Addr) & DJ_transfer_mark_temp_to_temp, u_in->val1, u_in->val2, 2);
            
            VG_(copy_UInstr)(cb, u_in);
            break;
         } // end case LEA1
         case LEA2: // DONE! *********************************************************
         {
            // const + TempReg1 + TempReg2 = TempReg3
            
            myAssert(u_in->tag1 == TempReg, "LEA2 tag1 is not TempReg when should be!");
            myAssert(u_in->tag2 == TempReg, "LEA2 tag2 is not TempReg when should be!");
            myAssert(u_in->tag3 == TempReg, "LEA2 tag3 is not TempReg when should be!");
            
            // just transfer the suppression mark from source to destination
            VG_(ccall_LLL_0) (cb, (Addr) & DJ_transfer_mark_temp_temp_to_temp, u_in->val1, u_in->val2, u_in->val3, 3);
            
            VG_(copy_UInstr)(cb, u_in);
            break;
         } // end case LEA2
         case INCEIP: // DONE! *********************************************************
         {
            // no need to do any suppression tracking stuff here
            
            // increment instance for curr x86 instruction
            x86InstrStructNode = getX86InstrInfo(curr_x86_addr);
            myAssert(x86InstrStructNode != NULL, "in case INCEIP, could not find x86InstrStruct in hashtable!");
            VG_(ccall_L_0) (cb, (Addr) & XZ_HELPER_increment_inst, (UInt)(x86InstrStructNode), 1);
         
            VG_(copy_UInstr)(cb, u_in);
            
            curr_x86_addr += u_in->val1;
            break;
         } // end case INCEIP         
         case GETSEG: // DONE! *********************************************************
         {
            // moves from an archReg (segment of it) to a tempReg
            myAssert(u_in->tag1 == ArchRegS, "GETSEG tag1 is not ArchRegS when should be!");
            myAssert(u_in->tag2 == TempReg, "GETSEG tag2 is not TempReg when should be!");
            
            // just transfer the suppression mark from source to destination
            VG_(ccall_LL_0) (cb, (Addr) & DJ_transfer_mark_archS_to_temp, u_in->val1, u_in->val2, 2);
            
            VG_(copy_UInstr)(cb, u_in);
            break;
         } // end case GETSEG
         case PUTSEG: // DONE! *********************************************************
         {
            // moves from a tempReg to an archReg (segment of it)
            myAssert(u_in->tag1 == TempReg, "PUTSEG tag1 is not TempReg when should be!");
            myAssert(u_in->tag2 == ArchRegS, "PUTSEG tag2 is not ArchRegS when should be!");
            
            // just transfer the suppression mark from source to destination
            VG_(ccall_LL_0) (cb, (Addr) & DJ_transfer_mark_temp_to_archS, u_in->val1, u_in->val2, 2);
            
            VG_(copy_UInstr)(cb, u_in);
            break;
         } // end case PUTSEG
         case USESEG: // DONE! *********************************************************
         {
            // uses from a tempReg to a tempReg (segments)
            myAssert(u_in->tag1 == TempReg, "USESEG tag1 is not TempReg when should be!");
            myAssert(u_in->tag2 == TempReg, "USESEG tag2 is not TempReg when should be!");
            
            // just transfer the suppression mark from source to destination
            VG_(ccall_LL_0) (cb, (Addr) & DJ_transfer_mark_temp_to_temp, u_in->val1, u_in->val2, 2);
            
            VG_(copy_UInstr)(cb, u_in);
            break;
         } // end case USESEG
         default: // DONE! *********************************************************
         {
            // just copy over the existing instruction
            VG_(copy_UInstr)(cb, u_in);
            break;
         }
      } // end big "switch" statement
      
   } // end "for" loop through each instruction

   VG_(free_UCodeBlock)(cb_in);
   return cb;
}



Bool gotToForcedInstance = False;


// Instrument the current basic block in PDG-CREATION mode! (written by Xiangyu, modified by Dennis)
UCodeBlock* SK_(instrument)(UCodeBlock* cb_in, Addr orig_addr)
{
   UCodeBlock* cb;
   Int         i,j,k,l, def, dj_i;
   UInstr*     u_in, * tmp_in;
   Char        fnname[128];
   Int line;
   fileCC * cur_fileCC;
   fnCC * cur_fnCC;
   fun_entry * fe;
   instr_info * ii=NULL, *cmp_ii=NULL;
   Bool is_fun_entry;
   Addr cur_x86_addr, prev_x86_addr = 0x0;
   Bool is_new_instr;
   Int r_tmp, r_dest, r_tmp1,r_tmp2;
   
   Int stack_init_idx=-1;
   Int predicate_idx=-1, pred_idx=-1;
   Int cur_uport=1;//0 is for control dependence
   Int internal_defs[MAX_INTERNAL_DEF];
   Int cur_def=0;
   Bool internal_defined;
   Int instr_regs[VG_MAX_REGS_USED];
   Bool instr_regs_iswrite[VG_MAX_REGS_USED];

   Bool stack_freed=False;
   instr_info * bb=NULL,*pred_ii=NULL,*start_bb;
   Bool after_get_ebp=False;
   Int stack_free_idx=-1;
   Int ebp_reg=-1;
   Int instr_reg_used;
   Int esp_reg=-1;
   Bool consider_defined; 
   Bool debug_terminate=False,jmpStamped=False;
   Int last_define_reg=-1, last_def_size=0; /* for REGISTER definitions */
   Int last_define_val=-1; /* for MEMORY definitions */
   Bool whetherFlagsWritten=False; /* tells whether FLAGS were written for this x86 instruction */
   
   FlagSet cmp_flags;
   
   Bool insideMainProgFile = False;
   Bool gotFPUInstruction = False;
   Int regToStoreFloatingVal = 1;
   Bool definedLocationIsFPUStack = False;
   instr_info *lastIIFromBasicBlock;
   
   // DENNIS variables for when we're doing force-mapping mode with trace-after-force turned on
   Bool instrumentHere = False;
   mapping_force **mappings = NULL;
   Int force_mode_archRegUse[7];
   Int force_mode_tempRegUse[7];
   Int force_mode_archRegDef[7];
   Int force_mode_tempRegDef[7];
   Int force_mode_nextUsePort;
   Int force_mode_nextDefPort;
   
   // DENNIS added this check: if we're in force-mapping mode (*without* trace-after-force flag),
   // then we do *not* need to
   // do all this extra instrumentation for values, etc.  The instrumented code should be
   // identical to the original code, *except* for when we're doing the forcing stuff.  So,
   // if we're in force-mapping mode with the trace-after-force flag *not* set,
   // we call *another* function to actually do the instrumentation
   if(valgrindMode == MODE_FORCE_MAPPING && doTraceAfterForce == False)
   {
      VG_(printf)("Error: this mode not currently supported\n");
      VG_(exit)(0);
   }
   
   // DENNIS added this check: if we're in stmt_coverage mode, then we call a separate function
   // to handle this case
   if(valgrindMode == MODE_STMT_COVERAGE)
   {
      VG_(printf)("Error: this mode not currently supported\n");
      VG_(exit)(0);
   }
   
   #ifdef OPTIMIZE_TRACE_AFTER_FORCE
   // DENNIS added this check: if we're in force-mapping mode (*with* trace-after-force), then
   // call another function to do this on-the-fly to limit what is traced to only the necessary
   // stuff (so the efficiency is improved).  Note: this will render lots of code below dead, since
   // some code below will also handle force-mapping mode with trace-after-force (it's just that
   // it traces much more stuff than is necessary!)
   if(valgrindMode == MODE_FORCE_MAPPING && doTraceAfterForce == True)
   {
      VG_(printf)("Error: this mode not currently supported\n");
      VG_(exit)(0);
   }
   #endif
   
   if(valgrindMode == MODE_SUPPRESS)
   {
      return DJ_instrument_suppress(cb_in, orig_addr);
   }
   
   
   
   
   cb = VG_(setup_UCodeBlock)(cb_in);
   
   // "STOCK" valgrind with no additional instrumentation
   for (i = 0; i < VG_(get_num_instrs)(cb_in); i++)
   {
      u_in = VG_(get_instr)(cb_in, i);

      switch(u_in->opcode)
      {
         default:
         {
            VG_(copy_UInstr)(cb, u_in);
            break;
         }
      }
   }
   VG_(free_UCodeBlock)(cb_in);
   return cb;










   
   
   VG_(printf)("Error: this mode not currently supported\n");
   VG_(exit)(0);
   
   
   //VG_(printf)("Start instrument\n");

   cb = VG_(setup_UCodeBlock)(cb_in);
   r_tmp=newTemp(cb);
   r_tmp1=SHADOW(r_tmp);
   r_tmp2=SHADOW1(r_tmp);
   
   //get debug information
   get_bb_info(orig_addr,&cur_fileCC, &cur_fnCC, &line);
   
#ifdef TRACK_ARGV
   //VG_(printf) ("CUR_FN=%s\n",cur_fnCC->fn);
   if (!main_entry && !VG_(strcmp)(cur_fnCC->fn,"main"))
   {
	   main_entry = 1;
      instrument_main_entry (cb,r_tmp,r_tmp1);
   }
#endif

   //if it is a new function, the control dependence is read in and the instructions for bb entries will be created with INVALID line and uport
   fe = get_fun_entry_new(orig_addr, cur_fileCC, cur_fnCC);

   if (fe->cd_available)
   {
      bb=search_for_bb(orig_addr, fe);
   }

   /* Count call to dlrr(), if this BB is dlrr()'s entry point */
   is_fun_entry = VG_(get_fnname_if_entry)(orig_addr, fnname, 128);

   //pred_idx is for tracing predicate results
   pred_idx = search_for_predicate(cb_in,0);

   if (fe->cd_available)
   {
	   predicate_idx=pred_idx;
   }
   if (fe->cd_available)
   {
      stack_free_idx=stack_free_space_find_index_new(cb_in,&esp_reg);
   }
   if (is_fun_entry && fe->cd_available)
   {
      stack_init_idx = stack_alloc_space_find_index_new(cb_in, fe);
   }

   cur_x86_addr = orig_addr;
   ii = NULL;
   ebp_reg = -1;
   
   // DENNIS added these so we can keep track of which ArchRegs are defined or used, and
   // what Temp regs correspond to them, since
   // this information needs to be outputted to the PDG file.
   // Note: for a given x86 instruction, there should be at most 7 used values (hardcoded by
   // Xiangyu as indicated by helper_reg_dep_ports array).  Assume the same for defined
   // values (although really, only one will be the *last* defined value that is actually
   // needed in the end).
   Int archRegUse[7];
   Int tempRegUse[7];
   Int archRegDef[7];
   Int tempRegDef[7];
   Int nextUsePort;
   Int nextDefPort;
   DJ_RESET_ARCH_REGS
   
   if(is_fun_entry)
   {
       // We're at the entry x86 instruction of a new function call!
       Int myTemp = newTemp(cb);
	   uInstr2(cb, GET, 4, ArchReg, R_ESP, TempReg, myTemp); 
       VG_(ccall_LR_0) (cb, DJ_got_func_entry, (UInt)(cur_fnCC), myTemp, 2);
   }
   
   // main loop through every valgrind instr in the basic block
   for (i = 0; i < VG_(get_num_instrs)(cb_in); i++)
   {
      consider_defined = False;

      u_in = VG_(get_instr)(cb_in, i);
      
      /* DENNIS stuff to continue working on later
      if(prev_x86_addr != cur_x86_addr)
      {
         prev_x86_addr = cur_x86_addr;
         VG_(ccall_0_0) (cb, (Addr) &DJ_junk);
      }
      */
      
      // if we're in force-mapping mode with trace-after-force turned on, then we need to check
      // whether or not we've gotten to a new x86 instruction that needs to be forced
      if(valgrindMode == MODE_FORCE_MAPPING && doTraceAfterForce == True)
      {
         VG_(printf)("Error: this mode not currently supported\n");
         VG_(exit)(0);
      } // end if we're in force-mapping mode with trace-after-force turned on
      
      

      if (i == stack_init_idx)
      {
	      stack_alloc_space_instrument_new(cb,u_in,fe);
	      ebp_reg = u_in->val1;
      }
      
      gotFPUInstruction = False;

      // instrument depending upon the current instruction type
      switch (u_in->opcode)
      {
         // DENNIS added this case "INC", so that the instruction is tracked.  The reason
         // is that sometimes, a JMP instruction is immediately preceeded by an "INC",
         // and in order to record the branch properly, the INC instruction must be tracked
         case INC:
         {
            TRACK_INSTR
            VG_(copy_UInstr)(cb, u_in);
            break;
         }
            
            
            
         // DENNIS added this case "DEC", so that the instruction is tracked.  The reason
         // is that sometimes, a JMP instruction is immediately preceeded by a "DEC",
         // and in order to record the branch properly, the DEC instruction must be tracked
         case DEC:
         {
            TRACK_INSTR
            VG_(copy_UInstr)(cb, u_in);
            break;
         }
      
      
      
         case GETSEG:
         {
            sk_assert(u_in->tag2 == TempReg);
	         TRACK_INSTR
            XZ_set_instr_inst(cb,ii,SHADOW(u_in->val2), SHADOW1(u_in->val2));
            VG_(copy_UInstr)(cb, u_in);
	         break;
	      }




         case INCEIP:
         {
	         if (ii)
	         {
		         SETUP_CD //this instruction has been tracked, thus CD need to establish
		         UPDATE_INSTR_UPORT //update the number of use ports for current instruction
		         INCREMENT_COUNTER
            }

#ifdef TIMESTAMPED
            // DENNIS changed the following so that a timestamp is *always* recorded when we 
            // process an INCEIP instruction
            /*
	         //TRACING control flow
	         if (bb && cur_x86_addr==bb->orig_addr)
	         { //trace timestamp only when it is the first ii for a bb
		         if (!ii)
		         { //ii is not set up for dependence
		            //INCREMENT_COUNTER
	               VG_(ccall_L_0) (cb, (Addr) & XZ_HELPER_increment_inst, (UInt) bb, 1);
		         }
	            VG_(ccall_L_0) (cb, (Addr) &XZ_HELPER_trace_timestamp, (UInt)bb, 1);
	         }
	         */

		      if(ii)
		      {
	            VG_(ccall_L_0) (cb, (Addr) &XZ_HELPER_trace_timestamp, (UInt)ii, 1);
	         }
#endif
            
#ifdef VALUE_WANTED
            recordDefinition(ii, last_define_reg, last_define_val, tempRegDef, archRegDef, cb,
                             last_def_size, whetherFlagsWritten, definedLocationIsFPUStack);
#endif // VALUE_WANTED

            if(instrumentHere == True)
	         {
	            // we must be at an x86 instruction to force (in force-mapping mode with trace-after-force turned on)
	            if(valgrindMode != MODE_FORCE_MAPPING || doTraceAfterForce == False)
	            {
	               VG_(printf)("Error: instrumentHere is TRUE when we're not in force-mapping mode with trace-after-force?!?!");
	               VG_(exit)(0);
	            }
	            
	            // do the force here!
	            DJ_instrument_force_instruction_inceip(cb, u_in, mappings, cur_x86_addr);
	            
	            // note: the above function call will copy over the inceip instruction
	            
	            // *repeat* getting the flags (if they were defined by this x86 instr) since
	            // the above function call may actually write new, forced flag values!
	            if(whetherFlagsWritten == True)
	            {
	               // get the defined flags
                  Int flagTempReg = 1; /* register t1 is not used elsewhere, so I'll use it here */
                  uInstr1(cb, GETF, 4, TempReg, flagTempReg);
                  uFlagsRWU(cb, FlagsALL, FlagsEmpty, FlagsEmpty); /* specify the previous GETF instruction should READ ALL FLAGS */	            
	               VG_(ccall_LR_0) (cb, DJ_set_flags_defined, (UInt)ii, flagTempReg,2);
	            }
	         }
            else
            {
               // copy over the INCEIP instruction
               VG_(copy_UInstr)(cb, u_in);
            }
            
            
            cur_x86_addr += u_in->val1;

            if (fe->cd_available)
            {
               bb = search_for_bb(cur_x86_addr, fe);
            }
            
	         last_define_reg=-1;
	         last_define_val=-1;
	         whetherFlagsWritten=False;
	         last_def_size=0;
            
            ii = NULL;
	         cur_uport = 1;
	         cur_def = 0;
	         after_get_ebp = False;
	         
	         definedLocationIsFPUStack = False;
	    
	         // DENNIS reset these variables so they're ready for the next x86 instruction
     	      DJ_RESET_ARCH_REGS
	    
            break;
         } // end case INCEIP




         case LOAD:
         {	
	         TRACK_INSTR
	         
	         if(instrumentHere == True)
	         {
	            // we must be at an x86 instruction to force (in force-mapping mode with trace-after-force turned on)
	            if(valgrindMode != MODE_FORCE_MAPPING || doTraceAfterForce == False)
	            {
	               VG_(printf)("Error: instrumentHere is TRUE when we're not in force-mapping mode with trace-after-force?!?!");
	               VG_(exit)(0);
	            }
	            
	            XZ_set_instr_inst(cb,ii,SHADOW(u_in->val2), SHADOW1(u_in->val2));
	            
	            // do the force here!
	            Int registerLoadedFrom = DJ_instrument_force_instruction_load(cb, u_in, mappings, cur_x86_addr,
	                                                              force_mode_archRegUse, force_mode_tempRegUse);
	            
	            // now do the regular value instrument stuff, given that we're loading from the
	            // register "registerLoadedFrom"
	            //if(!ignoreThisX86Instruction(u_in))
               //{
                  // update the arch/temp reg mappings so that the new reg registerLoadedFrom
                  // is also associated with the same arch reg as temp reg u_in->val1
                  for(dj_i = 0; dj_i < 7; dj_i++)
		            {
		               if(tempRegUse[dj_i] == u_in->val1)
		               {
		                  //tempRegUse[dj_i] = registerLoadedFrom;
		                  //break;
		                  archRegUse[nextUsePort] = archRegUse[dj_i];
                        tempRegUse[nextUsePort] = registerLoadedFrom;
                        nextUsePort++;
                        if(nextUsePort >= 7)
                        {
                           VG_(printf)("Error: more than 7 uses at a single x86 instruction?!?!\n");
                           VG_(exit)(0);
                        }
		                  
		                  break;
		               }
		            }
                  
                  MEM_DEP(registerLoadedFrom)
               //}
               
               // NOTE: do not need to repeat the load since it was already done above
	         }
	         else
	         {
	            // just do regular instrument without any value forcing
	         
	            // DENNIS commented out because in a load instruction, the register is an *address*
	            // from which to load, and we are not interested in this address value
	            // REG_DEP (u_in->val1)
       
               //if(!ignoreThisX86Instruction(u_in))
               //{
                  MEM_DEP (u_in->val1)
               //}

               XZ_set_instr_inst(cb,ii,SHADOW(u_in->val2), SHADOW1(u_in->val2));
               VG_(copy_UInstr)(cb, u_in);
               //DEBUG
               //#define DEBUG
#ifdef DEBUG
               //VG_(ccall_RR_0) (cb, (Addr) & XZ_HELPER_debug_load_value, u_in->val1, u_in->val2, 2);
#endif
            }
            
	         break;
	      } // end case LOAD




         case STORE:
         {
	         TRACK_INSTR
	    
	         if(instrumentHere == True)
	         {
	            // we must be at an x86 instruction to force (in force-mapping mode with trace-after-force turned on)
	            if(valgrindMode != MODE_FORCE_MAPPING || doTraceAfterForce == False)
	            {
	               VG_(printf)("Error: instrumentHere is TRUE when we're not in force-mapping mode with trace-after-force?!?!");
	               VG_(exit)(0);
	            }
	            
	            // do the force here!
	            Int sourceRegOfTheStore = DJ_instrument_force_instruction_store(cb, u_in, mappings, cur_x86_addr, cb_in, i);
	            
	            // now do the regular value instrument stuff
	            //if(!ignoreThisX86Instruction(u_in))
               //{
                  REG_DEP(u_in->val1) // dependence to same temp reg originally as the source for the store
               //}

               VG_(ccall_LLR_0) (cb, (Addr) &XZ_HELPER_store_mem_dep, (UInt)ii,u_in->size, u_in->val2,3);
               
               // NOTE: do not need to repeat the store since it was already done above
               
               last_define_reg = -1;
               last_define_val = sourceRegOfTheStore; // the last defined value is that of the source of the store
               last_def_size=u_in->size;
	         }
	         else
	         {
	            // just do regular instrument without any value forcing
	         
	            //if(!ignoreThisX86Instruction(u_in))
	            //{
	               REG_DEP(u_in->val1)
	      
	               // DENNIS commented out the following, because we do not want to output a dependence
	               // to the *address* of the memory location we're storing into
	               // REG_DEP(u_in->val2)
	            //}

	            VG_(ccall_LLR_0) (cb, (Addr) &XZ_HELPER_store_mem_dep, (UInt)ii,u_in->size, u_in->val2,3);

               //DEBUG
#ifdef DEBUG            
	            //VG_(ccall_LRR_0)(cb,(Addr) &XZ_HELPER_debug_store_value, cur_x86_addr, u_in->val2, u_in->val1,3);
#endif
               VG_(copy_UInstr)(cb, u_in);
                
               // DENNIS added this: we're storing value from reg "val1" into address "val2".  So this is
               // a *definition* using value "val1".  Keep track of this as being the last thing
               // defined so far!!!
               last_define_reg = -1;
               last_define_val = u_in->val1;
               last_def_size=u_in->size;
            }
            
            break;
         } // end case STORE
            
            
            

         // the following are not tracked
	      case GET:
	      {
	         if(instrumentHere == True)
	         {
	            // we must be at an x86 instruction to force (in force-mapping mode with trace-after-force turned on)
	            if(valgrindMode != MODE_FORCE_MAPPING || doTraceAfterForce == False)
	            {
	               VG_(printf)("Error: instrumentHere is TRUE when we're not in force-mapping mode with trace-after-force?!?!");
	               VG_(exit)(0);
	            }
	            
	            // do the force here!
	            DJ_instrument_force_instruction_get(cb, u_in, mappings, cur_x86_addr);
	            
	            // note: the above function does the get, and it's repeated below.  But it's okay to
	            // repeat the get instruction below since it doesn't hurt to get twice.
	         }
	      
            uInstr3(cb, GETV,4, ArchReg, u_in->val1,
	                 TempReg, SHADOW(u_in->val2),
	                 TempReg, SHADOW1(u_in->val2));
            //DEBUG
            //INSTRUMENT_DEBUG_GET
            VG_(copy_UInstr)(cb,u_in);
            /*
            VG_(ccall_LLR_0)(cb, (Addr) & XZ_HELPER_debug_print_reg,cur_x86_addr,u_in->val2,u_in->val2,3); 
	         if (u_in->val1==R_EBP)
	         {
	            //DEBUG
               //VG_(ccall_RL_0) (cb, (Addr) & XZ_HELPER_debug_EBP_get_value, u_in->val2,orig_addr, 1);
            }
            */
            
	         if(fe->cd_available && u_in->val1==R_EBP)
	         { 
	            //after_get_ebp=True;
		         ebp_reg = u_in->val2;
            }
       
            archRegUse[nextUsePort] = u_in->val1;
            tempRegUse[nextUsePort] = u_in->val2;
            nextUsePort++;
            if(nextUsePort >= 7)
            {
               VG_(printf)("Error: more than 7 uses at a single x86 instruction?!?!\n");
               VG_(exit)(0);
            }
       
	         break;
	      } // end case GET





	      case PUT:
	      {
	         if(instrumentHere == True)
	         {
	            // we must be at an x86 instruction to force (in force-mapping mode with trace-after-force turned on)
	            if(valgrindMode != MODE_FORCE_MAPPING || doTraceAfterForce == False)
	            {
	               VG_(printf)("Error: instrumentHere is TRUE when we're not in force-mapping mode with trace-after-force?!?!");
	               VG_(exit)(0);
	            }
	            
	            // do the force here!
	            DJ_instrument_force_instruction_put(cb, u_in, mappings, cur_x86_addr);
	            
	            // note: the above function does the put, and it's repeated below.  But it's okay to
	            // repeat the put instruction below since it doesn't hurt to put twice.
	         }
	      
            uInstr3(cb, PUTV,4, TempReg, SHADOW(u_in->val1),
		              TempReg, SHADOW1(u_in->val1),
                    ArchReg,u_in->val2);

	         // INSTRUMENT_DEBUG_PUT
            // VG_(ccall_LLR_0)(cb, (Addr) & XZ_HELPER_debug_print_reg,cur_x86_addr,u_in->val1,u_in->val1,3); 
	     
            VG_(copy_UInstr)(cb,u_in);
            
            /*
	         if(u_in->val2 == R_EBP)
	         {
               // DEBUG
               // VG_(ccall_RL_0) (cb, (Addr) & XZ_HELPER_debug_EBP_set_value, u_in->val1, orig_addr,1);
            }
            */

            // DENNIS added this
            archRegDef[nextDefPort] = u_in->val2;
            tempRegDef[nextDefPort] = u_in->val1;
            nextDefPort++;
            if(nextDefPort >= 7)
            {
               VG_(printf)("Error: more than 7 defs at a single x86 instruction?!?!\n");
               VG_(exit)(0);
            }

	         break;
	      } // end case PUT





         case MOV:
         {
            INSTRUMENT_OPERAND_SHADOW(u_in->tag1, u_in->val1, u_in->val2)
            VG_(copy_UInstr)(cb,u_in);
            
            // DENNIS added this: if the source of the move is a temp register (not a literal),
            // then look up the archReg associated with the source temp reg, and add a new
            // mapping that maps the same archReg to the destination temp reg
            if(u_in->tag1 == TempReg)
            {
               for(dj_i = 0; dj_i < 7; dj_i++)
               {
                  if(tempRegUse[dj_i] == u_in->val1)
                  {
                     archRegUse[nextUsePort] = archRegUse[dj_i];
                     tempRegUse[nextUsePort] = u_in->val2;
                     nextUsePort++;
                     if(nextUsePort >= 7)
                     {
                        VG_(printf)("Error: more than 7 uses at a single x86 instruction, huh?!?!\n");
                        VG_(exit)(0);
                     }
                     break;
                  }
               }
            }
            
            // DENNIS added this: if the source of the move is a literal, then it *could* be used
            // as an address for a subsequent load.  So, map the destination temp register to
            // a special value indicating that it's a load from a constant value address.
            if(u_in->tag1 == Literal)
            {
               archRegUse[nextUsePort] = LOCATION_MEM_CONST;
               tempRegUse[nextUsePort] = u_in->val2;
               nextUsePort++;
               if(nextUsePort >= 7)
               {
                  VG_(printf)("Error: more than 7 uses at a single x86 instruction, huh huh?!?!\n");
                  VG_(exit)(0);
               }
            }

            break;
         } // end case MOV
 




	      case LEA1:
	      {
	         INSTRUMENT_OPERAND_SHADOW(TempReg, u_in->val1, u_in->val2)
            VG_(copy_UInstr)(cb,u_in);
            OFFSET_EBP_ACCESS_LEA1
            
            // DENNIS added this: look up the archReg associated with the source temp reg, and add a new
            // mapping that maps the same archReg to the destination temp reg
            for(dj_i = 0; dj_i < 7; dj_i++)
            {
               if(tempRegUse[dj_i] == u_in->val1)
               {
                  archRegUse[nextUsePort] = archRegUse[dj_i];
                  tempRegUse[nextUsePort] = u_in->val2;
                  nextUsePort++;
                  if(nextUsePort >= 7)
                  {
                     VG_(printf)("Error: more than 7 uses at a single x86 instruction, huh what?!?!\n");
                     VG_(exit)(0);
                  }
                  break;
               }
            }
            
	         break;
	      } // end case LEA1






	      case LEA2:
	      {
	         TRACK_INSTR
	    
	         //if(!ignoreThisX86Instruction(u_in))
	         //{
	            REG_DEP(u_in->val1)
	            REG_DEP(u_in->val2)
	         //}
	    
            XZ_set_instr_inst(cb,ii,SHADOW(u_in->val3), SHADOW1(u_in->val3));
            VG_(copy_UInstr)(cb,u_in);
            OFFSET_EBP_ACCESS_LEA2
            
            // DENNIS added this: look up the archReg associated with the source temp reg 1 (of 2), and add a new
            // mapping that maps the same archReg to the destination temp reg.  Note that for LEA2, there are
            // *two* source registers, but we just need to save ONE of them to know at least one arch reg that
            // is related to the address being loaded from
            for(dj_i = 0; dj_i < 7; dj_i++)
            {
               if(tempRegUse[dj_i] == u_in->val1)
               {
                  archRegUse[nextUsePort] = archRegUse[dj_i];
                  tempRegUse[nextUsePort] = u_in->val3;
                  nextUsePort++;
                  if(nextUsePort >= 7)
                  {
                     VG_(printf)("Error: more than 7 uses at a single x86 instruction, huh wha?!?!\n");
                     VG_(exit)(0);
                  }
                  break;
               }
            }
            
	         break;
	      } // end case LEA2
	    
	    


	    
         /*
	      case RCL: case RCR: case SHL:
         case SHR: case SAR:
         case ROL: case ROR:
         {
	         INSTRUMENT_OPERAND_SHADOW(u_in->tag1, u_in->val1, u_in->val2)
            VG_(copy_UInstr)(cb,u_in);
	         break;
	      }
         */





         case CC2VAL:
         {
            if(instrumentHere == True)
	         {
	            // we must be at an x86 instruction to force (in force-mapping mode with trace-after-force turned on)
	            if(valgrindMode != MODE_FORCE_MAPPING || doTraceAfterForce == False)
	            {
	               VG_(printf)("Error: instrumentHere is TRUE when we're not in force-mapping mode with trace-after-force?!?!");
	               VG_(exit)(0);
	            }
	            
	            // do the force here!
	            DJ_instrument_force_instruction_cc2val(cb, u_in, mappings, cur_x86_addr);
	            
	            // note: the above function does the cc2val, and it's repeated below.  But it's okay to
	            // repeat the cc2val instruction below since it doesn't hurt to do it twice.
	         }
         
	         if (VG_IS_FLAG_OVERLAP(u_in->flags_r, cmp_flags))
	         {
	            TRACK_INSTR
	            ii->is_predicate = True;
	            
	            /* record the use if we're NOT in force mapping mode, or if we're in force-mapping mode AND in the main prog file */
	            if((valgrindMode != MODE_FORCE_MAPPING) || (valgrindMode == MODE_FORCE_MAPPING && insideMainProgFile == True))
	            {
	               VG_(ccall_LLL_0)(cb, (Addr) & XZ_HELPER_pred_dep, ii,cur_uport, cmp_ii,3);
	            }
	    
	            // note: the current instruction uses *flags* to determine whether to jump or not
               ii->usedRegister[cur_uport] = LOCATION_FLAG;
	    
	            cur_uport++;
               XZ_set_instr_inst(cb,ii,SHADOW(u_in->val1), SHADOW1(u_in->val1));
            }
            else
            {
               for (j=i-1;j>=0;j--)
               {
                  tmp_in=VG_(get_instr)(cb_in,j);
                  if (VG_IS_FLAG_OVERLAP(u_in->flags_r,tmp_in->flags_w))
                  {
                       break;
                  }
               }

               if(j >= 0)
               {
                  //found the comparison
                  //r_cd is the result temp reg to store the result bdd
                  if (tmp_in->tag3==TempReg)
                  {
		               INSTRUMENT_OPERAND_SHADOW(TempReg, tmp_in->val3, u_in->val1)
                  }
                  else if (tmp_in->tag2==TempReg)
                  {
         		      INSTRUMENT_OPERAND_SHADOW(TempReg, tmp_in->val2, u_in->val1)
                  }
                  else if (tmp_in->tag1==TempReg)
                  {
		               INSTRUMENT_OPERAND_SHADOW(TempReg, tmp_in->val1, u_in->val1)
                  }
                  else
                  {
	                  TRACK_INSTR
                     XZ_set_instr_inst(cb,ii,SHADOW(u_in->val1), SHADOW1(u_in->val1));
		            }
               }
               else
               {
	               TRACK_INSTR
                  XZ_set_instr_inst(cb,ii,SHADOW(u_in->val1), SHADOW1(u_in->val1));
               }
            }
            
            VG_(copy_UInstr)(cb, u_in);
            
            break;
         } // end case CC2VAL
            
            



	      case CMOV:
	      {
	         uInstr2(cb, CMOV, 4, TempReg, SHADOW(u_in->val1), TempReg, SHADOW(u_in->val2));
	         uInstr2(cb, CMOV, 4, TempReg, SHADOW1(u_in->val1), TempReg, SHADOW1(u_in->val2));
	         
            VG_(copy_UInstr)(cb, u_in);
            
            // DENNIS note: unlike for the MOV instruction, the CMOV instruction moves
            // a value to a temp register only if a condition holds.  Thus, the destination
            // temp register will always be gotten with a "GET" before the CMOV, since it's
            // possible that the condition will not hold and therefore the original value
            // should remain in that temp register (instead of the new value that was attempted
            // to be moved into it).  In other words, we DON'T need to adjust our internal
            // data structures here to map the destination temp register to an arch register,
            // since it is ALREADY mapped to an arch register.
            
            break;
         } // end case CMOV
	 
	 
	 
	 
	 
         case RCL:
         case RCR:
         case SHL:
         case SHR:
         case SAR:
         case ROL:
         case ROR:
         
         case ADD:
         case SUB:
         case ADC:
         case SBB:
         case MUL:
         case XOR:
         {  
            if(instrumentHere == True)
	         {
	            // we must be at an x86 instruction to force (in force-mapping mode with trace-after-force turned on)
	            if(valgrindMode != MODE_FORCE_MAPPING || doTraceAfterForce == False)
	            {
	               VG_(printf)("Error: instrumentHere is TRUE when we're not in force-mapping mode with trace-after-force?!?!");
	               VG_(exit)(0);
	            }
	            
	            // do the force here!
	            DJ_instrument_force_instruction_rcl_etc(cb, u_in, mappings, cur_x86_addr);
	            
	            // note: the above function does the instruction already.  So below, we only copy
	            // over the existing instruction if we *didn't* already do it above!
	         }
         
	         switch (u_in->tag1)
	         {  
 	            case TempReg:
 	            {
		            TRACK_INSTR
		            
	               //SETUP_TWO_EDGES
	               //if(!ignoreThisX86Instruction(u_in))
	               //{
                     REG_DEP(u_in->val1);
                     REG_DEP(u_in->val2);
                  //}
                  XZ_set_instr_inst(cb,ii,SHADOW(u_in->val2), SHADOW1(u_in->val2));
		            break;
		         }
	            case Literal:
	            {
	               // DENNIS added a TRACK_INSTR here, so that "cmp" instructions with a literal
	               // first argument are tracked, so that the jumps (branches) that use these values
	               // show up in the PDG file!!!!  Note: "cmp" in x86 translates to something with
	               // "sub" in it, in valgrind intermediate code
	               TRACK_INSTR
	            
	               // DENNIS did this check: if the first tag is a literal, then the *second* parameter
	               // is still a used value here!
	               //if(!ignoreThisX86Instruction(u_in))
	               //{
                     REG_DEP(u_in->val2);
                  //}
                  XZ_set_instr_inst(cb,ii,SHADOW(u_in->val2), SHADOW1(u_in->val2));
	            
	               break; 
	            }
	            case ArchReg: 
	            {
		            TRACK_INSTR
		         
		            archRegUse[nextUsePort] = u_in->val1;
                  tempRegUse[nextUsePort] = r_tmp;
                  nextUsePort++;
                  if(nextUsePort >= 7)
                  {
                     VG_(printf)("Error: more than 7 uses at a single x86 instruction here?!?!\n");
                     VG_(exit)(0);
                  }
		         
                  uInstr3(cb, GETV, 4, ArchReg, u_in->val1, TempReg, SHADOW(r_tmp), TempReg, SHADOW1(r_tmp)); 
               
                  //if(!ignoreThisX86Instruction(u_in))
                  //{
		            REG_DEP(r_tmp);
		            REG_DEP(u_in->val2);
		            //}

                  XZ_set_instr_inst(cb,ii,SHADOW(u_in->val2), SHADOW1(u_in->val2));
	               //SETUP_TWO_EDGES
	               break; 
	            }
            } // end switch on u_in->tag1
            
            if(instrumentHere == False)
            {
               // copy over the existing instruction
               VG_(copy_UInstr)(cb, u_in);
            }
            
            break;
         } // end case RCL, RCR, SHL, SHR, SAR, ROL, ROR, ADD, SUB, ADC, SBB, MUL, XOR





         case AND:
         case OR:
         {
	         TRACK_INSTR
	         //if(!ignoreThisX86Instruction(u_in))
	         //{
               REG_DEP(u_in->val1);
               REG_DEP(u_in->val2);
            //}
            XZ_set_instr_inst(cb,ii,SHADOW(u_in->val2), SHADOW1(u_in->val2));
	         //SETUP_TWO_EDGES
	         
            VG_(copy_UInstr)(cb, u_in);
            
            break;
         } // end case AND, OR





         case CALLM:
         {
	         TRACK_INSTR
            j = i-1;
	         tmp_in = VG_(get_instr) (cb_in,j);
            while (tmp_in->opcode!=CALLM_S)
            {
	            if (tmp_in->opcode == PUSH)
	            {
	               //if(!ignoreThisX86Instruction(u_in))
	               //{
		               REG_DEP(tmp_in->val1)
		            //}
		            if(cur_uport > 10)
		            {
		               VG_(exit)(0);
		            }
		         }
		         //VG_(printf) ("i=%d, j=%d, tmp_in->opcode=%d, CALLM_S=%d\n",i, j, tmp_in->opcode, CALLM_S);
		         j--;
               tmp_in = VG_(get_instr) (cb_in,j);
            } // end while loop
            
	         sk_assert(cur_uport<MAX_USE_PORT);

            VG_(copy_UInstr)(cb,u_in);
            j = i+1;
	         tmp_in = VG_(get_instr) (cb_in,j);
            while(tmp_in->opcode!=CALLM_E)
            {
	            if(tmp_in->opcode == POP)
	            {
                  XZ_set_instr_inst(cb,ii,SHADOW(tmp_in->val1), SHADOW1(tmp_in->val1));
               }
               j++;
               tmp_in = VG_(get_instr) (cb_in,j);
            } // end while loop

	         break;
	      } // end case CALLM





         case JMP:
         {
#ifdef INTER_CD
	         if (u_in->jmpkind==JmpCall)
	         {
		         if (!ii)
		         {
	               TRACK_INSTR
		         }
		
               uInstr2(cb, GET, 4, ArchReg, R_ESP, TempReg, r_tmp); 
	            VG_(ccall_LR_0) (cb, (Addr) &XZ_HELPER_create_dummy_CD, ii, r_tmp, 2);
		         // ii->cur_uport++;
	         }
#endif

	         if (pred_ii)
	         {
	            //if there is comparison, need to establish ii for jump and trace the value of jump
	            if (!ii)
	            {
	               TRACK_INSTR
               }

	            if (u_in->cond!=CondAlways)
	            {	
		            ii->is_predicate = True;
		            // VG_(printf)("JMP tracing %x, pred_ii=%x\n",ii->orig_addr,pred_ii->orig_addr);
		   
		            // establish the dependence between comparison and jump
		            
		            /* record the use if we're NOT in force mapping mode, or if we're in force-mapping mode AND in the main prog file */
	               if((valgrindMode != MODE_FORCE_MAPPING) || (valgrindMode == MODE_FORCE_MAPPING && insideMainProgFile == True))
	               {
		               VG_(ccall_LLL_0)(cb, (Addr) & XZ_HELPER_pred_dep, ii,cur_uport, pred_ii,3);
		            }
		    
		            // note: the current instruction uses *flags* to determine whether to jump or not
		            ii->usedRegister[cur_uport] = LOCATION_FLAG;
		    
		            cur_uport++;
		    
		            if (predicate_idx != -1)
		            {
                     if (bb->lv_idx==LV_NOT_ASSIGNED)
                     {	
                        //VG_(printf) ("NOT ASSIGNED bb is %x, bb->bb= %x\n",bb->orig_addr, bb->bb->orig_addr);
                        //goto CD_OUT;	     
                     }
                     else
                     {
                        instrument_predicate_with_ii(cb, cb_in, bb, ii, &r_tmp, & r_tmp1, &r_tmp2);	  
		               }
		            }

                  SETUP_CD
	               UPDATE_INSTR_UPORT
                  INCREMENT_COUNTER

#ifdef VALUE_WANTED
                  ii->definedRegister = LOCATION_JMP;

		          VG_(ccall_LL_0) (cb, (Addr) & XZ_HELPER_trace_value1, ii, 1, 2);
#endif

		         } // end if(u_in->cond != CondAlways)
		         else
		         {
		            //CondAlways
		            //Jcondition
		            //instrumentation for tracing 0
		            //jmp
		            
#ifdef VALUE_WANTED
                  ii->definedRegister = LOCATION_JMP;

		          VG_(ccall_LL_0) (cb, (Addr) & XZ_HELPER_trace_value1, ii, 0, 2); 
#endif
		         }

	         } // end if(pred_ii)
	    
	         else if(ii)
	         {
	            //otherwise , it is switch case statement / Jmp $EAX or Jmp-C
               SETUP_CD
	            UPDATE_INSTR_UPORT
               INCREMENT_COUNTER
            }

            // instrument for switch case
            if(fe->cd_available && predicate_idx == -1 && bb->lv_idx != LV_NOT_ASSIGNED && u_in->jmpkind != JmpRet)
            {
               //for switch case?

		         VG_(printf) ("WARNING, no predicate, but it is a CD\n");
               instrument_predicate(cb,cb_in,bb, u_in, &r_tmp, &r_tmp1);
            } 

#ifdef TIMESTAMPED
            // DENNIS altered this chunk of code to ensure that all jump instructions
            // have a timestamp associated with them
            /*
	         if (bb && u_in->jmpkind!=JmpCall && u_in->jmpkind!=JmpRet)
	         {
		         // basic block has only one instruction, no INCEIP
		         if (cur_x86_addr==bb->orig_addr && !jmpStamped)
		         {
		            jmpStamped=True;
 		            if (!ii)
 		            {
 		               //ii is not set up for dependence
		               //INCREMENT_COUNTER
	                  VG_(ccall_L_0) (cb, (Addr) & XZ_HELPER_increment_inst, (UInt) bb, 1);
		            }
	               VG_(ccall_L_0) (cb, (Addr) &XZ_HELPER_trace_timestamp, (UInt)bb, 1);
		         }
		      }
		      // if (u_in->jmpkind!=JmpCall && u_in->jmpkind!=JmpRet)
		      {
		         //basic block has only one instruction, no INCEIP
		      }
		      */
		      
		      if(ii)
		      {
	            VG_(ccall_L_0) (cb, (Addr) &XZ_HELPER_trace_timestamp, (UInt)ii, 1);
		      }

		      // the JCondition will be overwritten by Juncondition
		      if(bb)
		      {
		         if (u_in->tag1==Literal)
		         {
		            //VG_(ccall_LL_0) (cb, (Addr) & XZ_HELPER_trace_control_flow, bb, u_in->lit32,2);
		         }
		         else
		         {
		            //VG_(ccall_LR_0) (cb, (Addr) & XZ_HELPER_trace_control_flow, bb, u_in->val1,2);
	            }
	         }
#endif

            if(instrumentHere == True)
	         {
	            // we must be at an x86 instruction to force (in force-mapping mode with trace-after-force turned on)
	            if(valgrindMode != MODE_FORCE_MAPPING || doTraceAfterForce == False)
	            {
	               VG_(printf)("Error: instrumentHere is TRUE when we're not in force-mapping mode with trace-after-force?!?!");
	               VG_(exit)(0);
	            }
	            
	            // do the force here to ensure that we take the appropriate jump!
	            DJ_instrument_force_instruction_jmp(cb, u_in, mappings, cur_x86_addr, cb_in, i);
	            
	            // note: the above function call already does the jump
	         }
            else
            {
               // copy over the existing jump instruction to cause the jump to occur
               VG_(copy_UInstr)(cb, u_in);
            }
            
            ii = NULL;
	         cur_uport = 1;
	         cur_def = 0;
	         if (u_in->cond!=CondAlways &&bb&& bb->pred_ii)
	         {
		         bb->pred_ii->jmp_pc = cur_x86_addr;
 	         }
  	    
      	   // DENNIS reset these variables so they're ready for the next x86 instruction
     	      DJ_RESET_ARCH_REGS

            break;
         } // end case JMP
	   
	    
	    case FPU:
	    {
	       // The FPU floating-point instruction doesn't affect memory; I believe it only affects what is on
	       // the floating-point stack (which is not simulated by Valgrind), so we don't instrument anything
	       // here.
	       VG_(copy_UInstr)(cb, u_in);
	       break;
	    }
	    case FPU_R:
	    {  
	       TRACK_INSTR		   

		   // This instruction uses the lit16 encoding in val1 field to load the data from the address
		   // stored in the TempReg in the val2 field, onto the top of the floating-point stack.
		   // First, allow the load to happen as normal.
		   VG_(copy_UInstr)(cb, u_in);

		   // Now, we actually need to see what data was actually moved from memory onto the
		   // FPU stack.  To do this, we create a load instruction to load the data into a
		   // temp register, which will be considered the "defined value" for the current
		   // instruction
	       regToStoreFloatingVal = 1; // TempReg 1 should not be used here, so I'll use it
		   uInstr2(cb, LOAD, 4, TempReg, u_in->val2, TempReg, regToStoreFloatingVal);
		   
		   // now, instrument to record the memory dependence
		   //if(!ignoreThisX86Instruction(u_in))
           //{
              MEM_DEP (u_in->val2)
          // }
           XZ_set_instr_inst(cb,ii,SHADOW(u_in->val2), SHADOW1(u_in->val2));
               
           // set a flag so that we can later set the defined temp reg to be the
           // one holding the floating point value
           gotFPUInstruction = True; 
           
           // set a flag so that the defined location is marked as LOCATION_FPU
           definedLocationIsFPUStack = True;
               
	       break;
	    }
	    case FPU_W:
	    {
	       TRACK_INSTR		   

		   // This instruction uses the lit16 encoding in val1 field to store the data
		   // from the top of the floating-point stack into the address
		   // stored in the TempReg in the val2 field.
		   // First, allow the store to happen as normal.
		   VG_(copy_UInstr)(cb, u_in);

		   // Now, we actually need to see what data was actually moved from the FPU
		   // stack into memory.  To do this, we create a load instruction to load the data into a
		   // temp register, which will be considered the "defined value" for the current
		   // instruction
	       regToStoreFloatingVal = 1; // TempReg 1 should not be used here, so I'll use it
		   uInstr2(cb, LOAD, 4, TempReg, u_in->val2, TempReg, regToStoreFloatingVal);
		   
		   // now, instrument to record the memory dependence
           // (assume there are no uses since it's the top of the FPU stack that is really
           // being used here - the use doesn't matter, only the value moved from the FPU stack.
           // However, this could mess up slicing, so if we need to get slicing working later,
           // then the idea is to make this FPU_W be dependent upon the last-executed FPU_R
           // instruction, I think)
	       VG_(ccall_LLR_0) (cb, (Addr) &XZ_HELPER_store_mem_dep, (UInt)ii,u_in->size, u_in->val2,3);
               
           // set a flag so that we can later set the defined temp reg to be the
           // one holding the floating point value
           gotFPUInstruction = True; 
           
           // set a flag so that the defined location is marked as LOCATION_FPU
           definedLocationIsFPUStack = True;
	       
	       break;
	    }
	    
	    case DJ_RET:
	    {
	       VG_(ccall_L_0) (cb, (Addr) &DJ_got_func_ret, (UInt)(cur_fnCC), 1);
	       break;
	    }
	    /*
	    case DJ_LEAVE:
	    { TRACK_INSTR
           VG_(printf)("Got leave: %s\n", ii->fn);
	       break;
	    }
	    */
	    
         default:
         {
            // just copy over the existing instruction without any additional instrumentation
            VG_(copy_UInstr)(cb, u_in);
            break;
         }
            
      } // end the big "switch" for each instruction type

/*
      for (k=0;k<instr_reg_used;k++)
      {
	      if (instr_regs_iswrite[k])
	      {
	         // ebp_reg is redefined
	         if(ebp_reg==instr_regs[k])
	         {
	            ebp_reg = -1;
	            internal_defs[cur_def++] = instr_regs[k];
            }
         }
      }
*/


      if(i == pred_idx && ii)
      {
	      //ii->is_predicate = True;
	      pred_ii = ii;
	      if (predicate_idx!=-1)
	      {
	         bb->pred_ii = ii;
	      }
      }
     
      if(u_in->flags_w && ii)
      {
	      cmp_ii = ii;
	      cmp_flags = u_in->flags_w;
      }

      /*
      if (i==predicate_idx)
      {
	      //VG_(printf) ("i=%d, predicate_idx=%d\n",i,predicate_idx);
	      sk_assert(bb!=NULL);
	      //Bool need_to_increment=False;
	      if (!ii)
	      {
	         TRACK_INSTR
         }

	      //SETUP_CD
	      //UPDATE_INSTR_UPORT
         //need_to_increment=True;
          
	      bb->pred_ii = ii;
         if (bb->lv_idx==LV_NOT_ASSIGNED)
         {	
            VG_(printf) ("NOT ASSIGNED bb is %x, bb->bb= %x\n",bb->orig_addr, bb->bb->orig_addr);
            goto CD_OUT;	     
         }
         instrument_predicate(cb, cb_in, bb, u_in, &r_tmp, & r_tmp1);	        
      }

      if (debug_terminate)
      {
          INSTRUMENT_DEBUG_TERMINATE
      }
      */
      
      
CD_OUT:

      instr_reg_used = VG_(get_reg_usage) ( u_in, TempReg, instr_regs, instr_regs_iswrite);

      for (k=0; k < instr_reg_used; k++)
      {
	      if (instr_regs_iswrite[k])
	      {
	         // ebp_reg is redefined
	         if (u_in->opcode!=GET && ebp_reg==instr_regs[k])
	         {
	            ebp_reg = -1;
	         }
            if (consider_defined)
            {
	            internal_defs[cur_def++]=instr_regs[k];
		      }
		         
		      //trace value on last define
		      last_define_reg = instr_regs[k];
		      last_define_val = -1;
	  	      last_def_size = u_in->size;
         }
      }
      
      // special check for floating-point operations
      if(gotFPUInstruction == True)
      {
         last_define_reg = regToStoreFloatingVal;
      }

      // if the current ucode instruction wrote flags, then set a flag indicating this x86 instr wrote flags
      if(u_in->flags_w != FlagsEmpty)
      {
         whetherFlagsWritten = True;
      }        

      //pop %ebp
      if (i == stack_free_idx)
      {
	      //G_(printf)("XXXXX stack_free_idx=%d\n",i);
         if (u_in->opcode==INCEIP)
         {
            stack_free_space_instrument_new(cb, esp_reg,fe);
         }

         else
         {
	         stack_free_space_instrument_new(cb, u_in->val2, fe);
         }
	      stack_freed = True;
      }

   } // end the big "for" loop through each ucode instruction in the basic block


   // print_cb(cb);
   
   VG_(free_UCodeBlock)(cb_in);

   //VG_(printf)("End instrument\n");

   return cb;
}

heapAddrNode *heapAddrList = NULL;


void SK_(fini)(Int exitcode)
{
  Int i, j, k;
  mapping_force *mapping;

  Int pid=VG_(gettid)();
  
  if(valgrindMode == MODE_SUPPRESS)
  {
     printf("%d loads and %d stores\n", numLoads, numStores);
  }
  
  /*
  if(valgrindMode == MODE_PDG_CREATION)
  {
     // PDG file creation mode - need to write the PDG file (or the TRACE file, if
     // the pdg-to-trace flag is specified)
     if(doPdgToTrace == False)
     {
        // output the PDG data as-is
        
        Char buf[128];
        Char buf2[128];
        VG_(printf)("Writing PDG file for %d...\n",pid);

        // DENNIS commented out - instr_dic file not used right now
        // VG_(sprintf)(buf,"SLICING.%d.instr_dic",pid);
        // VG_(printf) ("finishing...\n");
        // write_instr_dic(buf);
        // VG_(printf) ("finishing... instr_dic done\n");
  
        VG_(sprintf)(buf,"TRACE.%d.pdg.txt",pid);
        VG_(sprintf)(buf2,"TRACE.%d.filter.txt",pid);
        write_dependence(buf, buf2);
        VG_(printf) ("PDG file created\n");
  
        
        //if (pid==open_tid)
        //{
           //VG_(printf)("closing output files for %d...\n",pid);
           //VG_(close)(output_fp);
           //VG_(close)(output_fp_pos);
        //}
        
     }
     else
     {
        // convert the PDG data to TRACE data and output it in that form
        
        Char buf[128];
        
        VG_(printf)("Writing TRACE file for %d...\n",pid);
        
        VG_(sprintf)(buf,"TRACE.%d.txt",pid);
        write_trace(buf, mappingsToForce, nextMappingToForceSlot, valgrindMode);
        
        VG_(printf) ("TRACE file created\n");
     } // end if the pdg-to-trace flag was specified or not!
   } // end if mode is PDG CREATION
   
   if(valgrindMode == MODE_FORCE_MAPPING)
   {
      // Sanity check: scan through the mapping_force structures and ensure that every
      // value to force was "MARKED TRUE" as having been forced during execution.
      // Warnings of mappings not forced are sent to a text file as well.
      
      // Note: it's possible that an instruction can have *multiple* of the SAME USED VALUES/LOCATIONS
      // (ex: "test %eax, %eax").  So, if a use has NOT been forced, then check whether another, identical
      // use at the same instruction has been forced.  Only if not, do we output the warning message.
      unlink("#vg_force_warnings.txt");
      for(i = 0; i < nextMappingToForceSlot; i++)
      {
         mapping = mappingsToForce[i];
      
         // special check: if the defined value is LOCATION_FPU, then assume it has been properly forced by
         // forcing the use!
         if(mapping->definedReg == LOCATION_FPU)
         {
            // ok, assume it was forced
         }
         
         // regular check: emit error if the defined value was not forced and the definedReg is NOT LOCATION_FPU
         else if(mapping->definedWasForced == False)
         {
            VG_(printf)("Warning: defined value at instr. 0x%x, instance #%d, was NOT FORCED!\n", mapping->instructionAddr, mapping->instanceNum);
            
            FILE *fp = fopen("#vg_force_warnings.txt", "a");
            fprintf(fp, "Warning: defined value at instr. 0x%x, instance #%d, was NOT FORCED!\n", mapping->instructionAddr, mapping->instanceNum);
            fclose(fp);
         }
         
         for(j = 0; j < mapping->numUses; j++)
         {
            if((mapping->usedWasForced)[j] == False)
            {
               // scan all *other* uses to see if there's a matching used value/location that was already forced
               Bool foundMatchingThatWasForced = False;
               for(k = 0; k < mapping->numUses; k++)
               {
                  if(k == j) continue;
                  
                  if((mapping->usedVals)[k] == (mapping->usedVals)[j] &&
                     (mapping->usedRegs)[k] == (mapping->usedRegs)[j] &&
                     (mapping->usedWasForced)[k] == True)
                  {
                     // found an identical use that *was* forced, so indeed, the current use was also actually
                     // forced as well!
                     foundMatchingThatWasForced = True;
                     break;
                  }
               }
               
               if(foundMatchingThatWasForced == False)
               {
                  VG_(printf)("Warning: used value at instr. 0x%x, instance #%d, used index %d, was NOT FORCED!\n", mapping->instructionAddr, mapping->instanceNum, j);
               
                  FILE *fp = fopen("#vg_force_warnings.txt", "a");
                  fprintf(fp, "Warning: used value at instr. 0x%x, instance #%d, used index %d, was NOT FORCED!\n", mapping->instructionAddr, mapping->instanceNum, j);
                  fclose(fp);
               }
            }
         }
      }
      
      // if the --trace-after-force flag was specified, we need to output
      // the tracing info (PDG file, or else the TRACE file, if --pdg-to-trace flag is
      // specified).
      if(doTraceAfterForce == True)
      {
         // in force-mapping mode, a timeout might happen due to an infinite loop, meaning
         // there could be a LOT of data, so we don't output in that case
         if(VG_(didValgrindTimeout)() == False)
         {
            if(doPdgToTrace == False)
            {
               // output the PDG data as-is
               Char buf[128];
               Char buf2[128];
               VG_(printf)("Writing PDG file for %d...\n",pid);
  
               VG_(sprintf)(buf,"TRACE.%d.pdg.txt",pid);
               VG_(sprintf)(buf2,"TRACE.%d.filter.txt",pid);
               write_dependence(buf, buf2);
               VG_(printf) ("PDG file created\n");
            }
            else
            {
               // convert the PDG data to TRACE data and output it in that form
               Char buf[128];
               VG_(printf)("Writing TRACE file for %d...\n",pid);
        
               VG_(sprintf)(buf,"TRACE.%d.txt",pid);

               #ifdef OPTIMIZE_TRACE_AFTER_FORCE
               write_trace_optimized(buf);
               #else
               write_trace(buf, mappingsToForce, nextMappingToForceSlot, valgrindMode);
               #endif
               
               VG_(printf) ("TRACE file created\n");
               
            } // end if the pdg-to-trace flag was specified or not!
         } // end if valgrind did NOT hit the timeout
      } // end if doTraceAfterForce == TRUE
      
      // free the mapping_force structures that were allocated, and the
      // force_helper_params structures that were allocated
      for(i = 0; i < nextMappingToForceSlot; i++)
      {
         if(mappingsToForce[i]->funcName != NULL)
         {
            free(mappingsToForce[i]->funcName);
         }
         if(mappingsToForce[i]->usedVals != NULL)
         {
            free(mappingsToForce[i]->usedVals);
         }
         if(mappingsToForce[i]->usedRegs != NULL)
         {
            free(mappingsToForce[i]->usedRegs);
         }
         if(mappingsToForce[i]->usedWasForced != NULL)
         {
            free(mappingsToForce[i]->usedWasForced);
         }
         if(mappingsToForce[i]->usedLocationWasReturned != NULL)
         {
            free(mappingsToForce[i]->usedLocationWasReturned);
         }
         if(mappingsToForce[i]->usedAddrType != NULL)
         {
            free(mappingsToForce[i]->usedAddrType);
         }
         if(mappingsToForce[i]->heapInstrs != NULL)
         {
            free(mappingsToForce[i]->heapInstrs);
         }
         if(mappingsToForce[i]->chunkNums != NULL)
         {
            free(mappingsToForce[i]->chunkNums);
         }
         if(mappingsToForce[i]->funcNames != NULL)
         {
            for(j = 0; j < mappingsToForce[i]->numUses; j++)
            {
               if((mappingsToForce[i]->funcNames)[j] != NULL)
               {
                  free((mappingsToForce[i]->funcNames)[j]);
               }
            }
            free(mappingsToForce[i]->funcNames);
         }
         if(mappingsToForce[i]->funcInsts != NULL)
         {
            free(mappingsToForce[i]->funcInsts);
         }
         
         free(mappingsToForce[i]);
      }
      for(i = 0; i < nextForceHelperParamSlot; i++)
      {
         free(forceHelperParams[i]);
      }
   } // end if valgrindMode == MODE_FORCE_MAPPING
   
   if(valgrindMode == MODE_STMT_COVERAGE)
   {
      // write out the covered stmts
      Char buf[128];
      VG_(printf)("Writing STMT COVERAGE file for %d...\n",pid);
      VG_(sprintf)(buf,"STMT_COV.%d.txt",pid);
      write_stmt_coverage(buf);
      VG_(printf) ("STMT COVERAGE file created\n");
   }
   
   // write the pid to an output file that will always hold the pid of the last-executed
   // valgrind run (if control reaches here, then that means valgrind executed successfully without
   // any errors that would have caused early program termination
   if(valgrindMode != MODE_SUPPRESS)
   {
      FILE *filePointer = fopen("#vg_last_pid.txt", "w");
      fprintf(filePointer, "%d\n", pid);
      fclose(filePointer);
   }
   */
   VG_(printf)("Program execution complete!\n");
}

/*
SK_(new_mem_stack)(Addr a, UInt len)
{
   VG_(printf)("0x%x(%u)\n", a, len);
}
*/

//___NR_write
#define SYSCALL_READ 3
#define SYSCALL_WRITE 4
#define SYSCALL_EXECVE 11

void * SK_(pre_syscall) (ThreadId tid, UInt syscallno, Bool isBlocking)
{
  Int fd, i,j,k;
  Char * buf, *buf2;
  Char ** argv;
  size_t count;
  void * tst;
  static Int total_count=0;
  Char t_buf[1024];
  Char c_buf[1024];
  instr_info * ii;
  Int inst;
  Char c;

  //if (0) {
  Int pid=VG_(gettid)();
  if (syscallno==SYSCALL_WRITE) {

        //VG_(printf) ("XXXXX SYS_WRITE (tid=%d):\n",pid);
        tst=VG_(XZ_get_threadstate_3_arg)(tid, &fd, &buf, &count);
        //VG_(printf) ("\tXXXXX SYS_WRITE good 1 count=%d,fd=%d, buf=%x\n",count,fd,buf);
        //VG_(printf) ("\tXXXXX SYS_WRITE buf[0]=%c, buf[15]=%c, count=%d\n",buf[0],buf[15],count);
	if (buf) {

      //==============================
      // DENNIS added this stuff to capture the output in a text file (only for certain valgrind modes)
      if(/*valgrindMode == MODE_PDG_CREATION || */valgrindMode == MODE_FORCE_MAPPING)
      {
	      //VG_(printf)("OUTPUT %s, length %d bytes\n", buf, count);
	      buf2 = VG_(malloc)((count*sizeof(char))+1);
	      VG_(memset)(buf2, '\0', count+1);
	      VG_(strncpy)(buf2,buf,count);
	      //VG_(printf)("OUTPUT %s\n", buf2);
   
         // check the last instruction executed from the main source code file
         ii = lastInstrExecutedFromMainSourceFile;
         //VG_(printf)("instance %d, addr 0x%x, file %s, function %s, line %d\n", lastInstrInstanceExecutedFromMainSourceFile, ii->orig_addr, ii->file, ii->fn, ii->line);

         // write the output information to the output file
         Int pid=VG_(gettid)();
         Char buffy[128];
         VG_(sprintf)(buffy,"TRACE.%d.output.txt",pid);
         FILE *filePointer = fopen(buffy, "a");
      
         if(ii == NULL)
         {
            fprintf(filePointer, "[OUTPUT_PART_BEGIN_%d %X#%d@%d]", nextOutputPart, 0x0, lastInstrInstanceExecutedFromMainSourceFile, 0);
         }
         else
         {
            fprintf(filePointer, "[OUTPUT_PART_BEGIN_%d %X#%d@%d]", nextOutputPart, ii->orig_addr, lastInstrInstanceExecutedFromMainSourceFile, ii->line);
         }
         fprintf(filePointer, "%s", buf2);
         fprintf(filePointer, "[OUTPUT_PART_END_%d]\n", nextOutputPart);
      
         nextOutputPart++;
      
         fclose(filePointer);
      }
   }
   }
   
      // END DENNIS stuff
      //==============================
      
      /*
	   //VG_(printf) ("SYS_WRITE:%s\n",buf);
	   //VG_(printf) ("\tCOUNT=%d\n",count);
	   VG_(sprintf) (t_buf, "(PID:%d) from %d, count=%d\n",pid,total_count,count);
	   //VG_(write) (output_fp, (void*) t_buf, VG_(strlen)(t_buf));
	   //VG_(write) (output_fp, (void*) buf,count);
	   VG_(write) (output_fp_pos, (void*) t_buf, VG_(strlen)(t_buf));
	   VG_(write) (output_fp_sim, (void*) t_buf, VG_(strlen)(t_buf));
	                //VG_(printf) ("OUTPUT %s\n",buf);
/*
           for (i=0;i<count;i++) {
	       c=buf[i];
	       VG_(sprintf)(t_buf,"%c",c);
               VG_(write)(output_fp_pos,(void*) t_buf, VG_(strlen)(t_buf));
           }
*
           k=0;  
           for (i=0;i<count;i++) {
               ii=(instr_info*) get_ii_vbytes4_ALIGNED((Addr)(buf+i));
               inst=get_inst_vbytes4_ALIGNED((Addr)(buf+i));
	       c=buf[i];
	       
	       if (!c) c='0';
	       if ((UInt)ii!=0xffffffff && ii ) {
		   VG_(sprintf) (t_buf,"&* %d,%d *&'%c'\n",ii->id,inst,c);
	           VG_(write) (output_fp_sim, (void*) t_buf, VG_(strlen)(t_buf));
		   VG_(sprintf) (t_buf,"%c(%d,%d)",c,ii->id,inst);
	       } else {
		   VG_(sprintf) (t_buf,"%c",c);

               }
	       //VG_(printf)("%s\n",t_buf);
	       VG_(write) (output_fp_pos, (void*) t_buf, VG_(strlen)(t_buf));
	       c_buf[k++]=c;
	       if (c=='\n') {
		   c_buf[k]='\0';
		   VG_(write) (output_fp_sim, (void*) c_buf, VG_(strlen)(c_buf));
		   k=0;
	       }
           }
	   
	   total_count+=count;
	   //VG_(printf)("SYS_WRITE done\n");
        }
  } else if (syscallno==SYSCALL_EXECVE) {
        //VG_(printf) ("XXXXX SYS_EXECVE (tid=%d):\n",pid);
        tst=VG_(XZ_get_threadstate_exec_argv)(tid, &argv);
      
        //VG_(printf) ("\tXXXXX SYS_WRITE good 1 count=%d,fd=%d, buf=%x\n",count,fd,buf);
        //VG_(printf) ("\tXXXXX SYS_WRITE buf[0]=%c\n",buf[0]);
	if (argv) {
	   //VG_(printf) ("SYS_WRITE:%s\n",buf);
	   //VG_(printf) ("\tCOUNT=EXECVE\n",count);
	   VG_(sprintf) (t_buf, "(PID:%d) from %d, count=%d\n",pid,total_count,count);
	   VG_(write) (output_fp, (void*) t_buf, VG_(strlen)(t_buf));
	   VG_(write) (output_fp, (void*) buf,count);
	   VG_(write) (output_fp_pos, (void*) t_buf, VG_(strlen)(t_buf));
	   VG_(write) (output_fp_sim, (void*) t_buf, VG_(strlen)(t_buf));
	                //VG_(printf) ("OUTPUT %s\n",buf);
           i=0;
           while (argv[i]) {
	      buf=argv[i];
              for (j=0;j<VG_(strlen)(argv[i]);j++) {
                  ii=(instr_info*) get_ii_vbytes4_ALIGNED((Addr)(buf+j));
                  inst=get_inst_vbytes4_ALIGNED((Addr)(buf+j));
	          if ((UInt)ii!=0xffffffff && ii) {
		     VG_(sprintf) (t_buf,"&* %d,%d *&'%c'\n",ii->id,inst,buf[j]);
	             VG_(write) (output_fp_sim, (void*) t_buf, VG_(strlen)(t_buf));
		     VG_(sprintf) (t_buf,"%c(%d,%d)",buf[j],ii->id,inst);
	          } else {
		     VG_(sprintf) (t_buf,"%c",buf[j]);
                  }
	          VG_(write) (output_fp_pos, (void*) t_buf, VG_(strlen)(t_buf));
	      }//for
	      i++;
           }
	   
	   total_count+=count;
	   //VG_(printf)("SYS_WRITE done\n");
        }
/*
        i=0;
	while (argv[i]) {
	   VG_(printf) ("argv[%d]=%s\n",i,argv[i]);
	   i++;
        }
*
  }
  //VG_(printf) ("XXXXX SYS_WRITE (tid=%d) DONE:\n",pid);
  */
  
  // DENNIS capturing syscall
  /*
  ii = lastInstrExecutedFromMainSourceFile;
  if(ii != NULL)
  {
     VG_(printf)("*** Instr 0x%x, system call #%d\n", ii->orig_addr, syscallno);
  }
  else
  {
     VG_(printf)("*** System call #%d\n", syscallno);
  }
  */
  
  return (void*) True;
}

__inline__ void  SK_(post_syscall) ( ThreadId tid, UInt syscallno,
                           void* pre_result, Int res, Bool isBlocking )
{
#ifdef TRACK_INPUT
  Int fd, i,j;
  Char * buf;
  Char ** argv;
  size_t count;
  void * tst;
  static Int total_count=0;
  Char t_buf[1024];
  instr_info * ii;
  Int inst;
  Char c;

  //if (0) {
  Int pid=VG_(gettid)();
  if (syscallno==SYSCALL_READ) {

        //VG_(printf) ("XXXXX SYS_READ (tid=%d):\n",pid);
        tst=VG_(XZ_get_threadstate_3_arg)(tid, &fd, &buf, &count);
        //VG_(printf) ("\tXXXXX SYS_WRITE good 1 count=%d,fd=%d, buf=%x\n",count,fd,buf);
        //VG_(printf) ("\tXXXXX SYS_WRITE buf[0]=%c, buf[15]=%c\n",buf[0],buf[15]);
	if (buf&& res) {
/*
	   VG_(printf) ("SYS_WRITE:%s\n",buf);
	   VG_(printf) ("\tCOUNT=%d\n",count);
	   VG_(sprintf) (t_buf, "(PID:%d) from %d, count=%d\n",pid,total_count,count);
*/
	   //VG_(printf) ("READ(%d)\n",res);
	   //VG_(printf) ("READ(%d)=%s\n",res,buf);
           for (i=0;i<res;i++) {
               set_ii_vbytes4_ALIGNED((Addr)(buf+i),INPUT_RELEVANT);
	       //VG_(printf) ("set '%c' at %x\n",buf[i], buf+i);
               set_inst_vbytes4_ALIGNED((Addr)(buf+i),buf[i]);
           }
	 }
   }
   
   // DENNIS capturing result of system call
   //VG_(printf)("System call #%d, result is %d\n", syscallno, res);
   
	   //VG_(write) (outp
     //VG_(printf) ("POST good\n");
#endif
}


VG_DETERMINE_INTERFACE_VERSION(SK_(pre_clo_init), 2/1)
//VG_DETERMINE_INTERFACE_VERSION(SK_(pre_clo_init), 4./8)


/*--------------------------------------------------------------------*/
/*--- end                                                lk_main.c ---*/
/*--------------------------------------------------------------------*/

